#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import numpy as np

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.set_cell_value_code_chunk import \
    SetCellValueCodeChunk
from mitosheet.errors import make_cast_value_to_type_error
from mitosheet.is_type_utils import (is_none_type, is_number_dtype,
                                     is_string_dtype)
from mitosheet.public.v1.sheet_functions.types import \
    get_function_to_convert_to_series
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import ColumnID, StepType


class SetCellValueStepPerformer(StepPerformer):
    """
    A set_cell_value step, allows you to set the value
    of a given cell in the sheet.

    NOTE: this step performer is no longer generated by the frontend, so we don't worry
    about making it work with new public interface versions.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'set_cell_value'

    @classmethod
    def saturate(cls, prev_state: State, params: Dict[str, Any], previous_steps: List[StepType]) -> Dict[str, Any]:
        # Mito doesn't allow empty cells, so if the new value is empty, change it to None.
        if params['new_value'] == '':
            params['new_value'] = None

        # Get the old value so we can check if the new value is different
        sheet_index = params['sheet_index']
        column_id = params['column_id']
        row_index = params['row_index']
        column_header = prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)

        # Cast the old value to a string to avoid errors while writing the saved analysis
        params['old_value'] = str(prev_state.dfs[sheet_index].at[row_index, column_header])
        
        return params

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        column_id: ColumnID = get_param(params, 'column_id')
        row_index: int = get_param(params, 'row_index')
        old_value: str = get_param(params, 'old_value')
        new_value: Union[str, None] = get_param(params, 'new_value')

        # If nothings changed, there's no work to do
        if old_value == new_value:
            return prev_state, None

        column_header = prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)

        # Update the value of the cell, we handle it differently depending on the type of the column
        column_dtype = str(prev_state.dfs[sheet_index][column_header].dtype)
        type_corrected_new_value = cast_value_to_type(new_value, column_dtype)

        execution_data = {
            'type_corrected_new_value': type_corrected_new_value,
        }

        return cls.execute_through_transpile(
            prev_state, 
            params, 
            execution_data,
        )

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:

        return [
            SetCellValueCodeChunk(
                prev_state, 
                get_param(params, 'sheet_index'), 
                get_param(params, 'column_id'), 
                get_param(params, 'row_index'), 
                get_param(params, 'old_value'), 
                get_param(params, 'new_value'), 
                execution_data['type_corrected_new_value'] if (execution_data is not None and 'type_corrected_new_value' in execution_data) else None)
        ]


    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}


def cast_value_to_type(value: Union[str, None], column_dtype: str) -> Optional[Any]:
    """
    Helper function for converting a value into the correct type for the 
    series that it is going to be added to. 
    """
    # If the user is trying to make the value None, let them.
    if is_none_type(value) or value is None:
        return None

    try:
        conversion_function = get_function_to_convert_to_series(column_dtype)
        casted_value_series = conversion_function(value, on_uncastable_arg_element=np.nan)

        type_corrected_new_value = casted_value_series.iat[0]

        # If the value is a string and it has a " in it, replace it with a ' so the transpiled code does not error
        if is_string_dtype(column_dtype) and '"' in type_corrected_new_value:
            type_corrected_new_value = type_corrected_new_value.replace('"', "'")

        # If the typed value is not a float, then we do not make it one
        if is_number_dtype(column_dtype) and '.' not in value:
            return round(type_corrected_new_value)

        return type_corrected_new_value
    except:
        raise make_cast_value_to_type_error(value, column_dtype, error_modal=False)



from typing import Any, Dict, Optional
from mitosheet.mito_backend import MitoBackend
from mitosheet.mito_flask.v1.flatten_utils import (flatten_mito_backend_to_json, read_backend_state_string_to_mito_backend)

def process_mito_event(backend_state: Optional[str], mito_event: Optional[Dict[str, Any]]) -> Any:

    if backend_state is None:
        mito_backend = MitoBackend()
        from flask import jsonify
        return jsonify({
            "state": flatten_mito_backend_to_json(mito_backend),
            "response": None,
        })

    mito_backend = read_backend_state_string_to_mito_backend(backend_state)

    response = None
    def mito_send(message):
        nonlocal response
        response = message
        
    mito_backend.mito_send = mito_send
    if mito_event:
        mito_backend.receive_message(mito_event)

    from flask import jsonify
    return jsonify({
        "state": flatten_mito_backend_to_json(mito_backend),
        "response": response,
    })


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import csv
import os
from os.path import basename, normpath
from typing import Any, Dict, List, Optional, Set, Tuple

import chardet
import pandas as pd

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.import_steps.simple_import_code_chunk import (
    DEFAULT_DECIMAL, DEFAULT_DELIMITER, DEFAULT_ENCODING,
    DEFAULT_ERROR_BAD_LINES, DEFAULT_SKIPROWS, SimpleImportCodeChunk)
from mitosheet.errors import (make_file_not_found_error,
                              make_invalid_simple_import_error,
                              make_is_directory_error)
from mitosheet.state import DATAFRAME_SOURCE_IMPORTED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.utils import get_valid_dataframe_names


class SimpleImportStepPerformer(StepPerformer):
    """
    A simple import, which allows you to import csv files 
    with the given file_names, while detecting the correct
    way to import them.
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'simple_import'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        file_names: List[str] = get_param(params, 'file_names')
        delimeters: Optional[List[str]] = get_param(params, 'delimeters')
        encodings: Optional[List[str]] = get_param(params, 'encodings')
        decimals: Optional[List[str]] = get_param(params, 'decimals')
        skiprows: Optional[List[int]] = get_param(params, 'skiprows')
        error_bad_lines: Optional[List[bool]] = get_param(params, 'error_bad_lines')
        use_deprecated_id_algorithm: bool = get_param(params, 'use_deprecated_id_algorithm') if get_param(params, 'use_deprecated_id_algorithm') else False

        # If any of the files are directories, we throw an error to let
        # the user know
        for file_name in file_names:
            if os.path.isdir(file_name):
                raise make_is_directory_error(file_name)
            if not os.path.exists(file_name):
                raise make_file_not_found_error(file_name)

        # Create a new step
        post_state = prev_state.copy()

        file_delimeters = []
        file_encodings = []
        file_decimals = []
        file_skiprows = []
        file_error_bad_lines = []
        new_df_names = []

        just_final_file_names = [basename(normpath(file_name)) for file_name in file_names]

        pandas_processing_time = 0.0
        for index, (file_name, df_name) in enumerate(zip(file_names, get_valid_dataframe_names(post_state.df_names, just_final_file_names))):
            
            # We try to read the csv with the parameters that the user specified. 
            # If the user has not specified parameters, then its because they did not go to the csv configure page, and instead
            # are using Mito's defaults. In this case, we're able to make educated guesses for the delimiter and encoding, and use pandas defaults
            # for the remainder of the parameters. 
            if delimeters is not None and encodings is not None:
                # Given the Mito UI, we expect that if the user has specified the delimiter and ecoding, 
                # that the rest of the parameters are also defined. The only time that is not the case is when 
                # the user is replaying an old simple import that does not have these fields. To account for that, 
                # we just handle the None case here. This makes it easy to add new parameters without having to write 
                # step upgraders. 
                # This approach of handling optional step params instead of writing a step upgrader is also used in graphs.
                delimeter = delimeters[index]
                encoding = encodings[index]
            else:
                _, delimeter, encoding = read_csv_get_delimiter_and_encoding(file_name)
                
            decimal = decimals[index] if decimals is not None else DEFAULT_DECIMAL
            _skiprows = skiprows[index] if skiprows is not None else DEFAULT_SKIPROWS
            _error_bad_lines = error_bad_lines[index] if error_bad_lines is not None else DEFAULT_ERROR_BAD_LINES
                

            # Save the delimeter and encodings for transpiling
            file_delimeters.append(delimeter)
            file_encodings.append(encoding)
            file_decimals.append(decimal)
            file_skiprows.append(_skiprows)
            file_error_bad_lines.append(_error_bad_lines)
            
            new_df_names.append(df_name)

        execution_data = {
            'file_delimeters': file_delimeters,
            'file_encodings': file_encodings,
            'file_decimals': file_decimals,
            'file_skiprows': file_skiprows,
            'file_error_bad_lines': file_error_bad_lines,
            'pandas_processing_time': pandas_processing_time,
            'new_df_names': new_df_names
        }

        try:
            return cls.execute_through_transpile(
                prev_state,
                params,
                execution_data,
                new_dataframe_params={
                    'df_source': DATAFRAME_SOURCE_IMPORTED,
                    'new_df_names': new_df_names,
                    'overwrite': None
                },
                use_deprecated_id_algorithm=use_deprecated_id_algorithm
            )
        except:
            raise make_invalid_simple_import_error()

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            SimpleImportCodeChunk(
                prev_state, 
                get_param(params, 'file_names'), 
                get_param(execution_data if execution_data is not None else {}, 'file_delimeters'), 
                get_param(execution_data if execution_data is not None else {}, 'file_encodings'), 
                get_param(execution_data if execution_data is not None else {}, 'file_decimals'), 
                get_param(execution_data if execution_data is not None else {}, 'file_skiprows'), 
                get_param(execution_data if execution_data is not None else {}, 'file_error_bad_lines'),
                get_param(execution_data if execution_data is not None else {}, 'new_df_names')
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}


def read_csv_get_delimiter_and_encoding(file_name: str) -> Tuple[pd.DataFrame, str, str]:
    """
    Given a file_name, will read in the file as a CSV, and
    return the df, delimeter, decimal separator, and encoding of the file
    """
    encoding = DEFAULT_ENCODING
    delimeter = DEFAULT_DELIMITER

    if is_url_to_file(file_name):
        df = pd.read_csv(file_name)
        return df, delimeter, encoding

    try:
        # First attempt to read csv without specifying an encoding, just with a delimeter
        delimeter = guess_delimeter(file_name)
        df = pd.read_csv(file_name, sep=delimeter)
    except UnicodeDecodeError:
        # If we have an encoding error, try and get the encoding
        try: 
            encoding = guess_encoding(file_name)
            delimeter = guess_delimeter(file_name, encoding=encoding)

            # Read the file as dataframe 
            df = pd.read_csv(file_name, sep=delimeter, encoding=encoding)
        except: 
            # Sometimes guess_encoding, guesses 'ascii' when we want 'latin-1', 
            # so if guess_encoding fails, we try latin-1
            encoding = 'latin-1'
            df = pd.read_csv(file_name, sep=delimeter, encoding=encoding)
        
    return df, delimeter, encoding


def guess_delimeter(file_name: str, encoding: Optional[str]=None) -> str:
    """
    Given a path to a file that is assumed to exist and be a CSV, this
    function guesses the delimeter that is used by that file
    """
    s = csv.Sniffer()
    with open(file_name, 'r', encoding=encoding) as f:
        return s.sniff(f.readline()).delimiter

def guess_encoding(file_name: str) -> str:
    """
    Uses chardet to guess the encoding of the the file
    at the given file_name
    """
    # Attempt to determine the encoding and try again. 
    with open(file_name, 'rb') as f:
        result = chardet.detect(f.readline())
        return result['encoding']


def is_url_to_file(file_name: str) -> bool:
    """
    Returns true if the file_name is a url to a file -- which we need to know
    as we can't sniff the delimeter of a url.

    This isn't perfect, and we should improve it, but it's ok for now!
    """
    if file_name.startswith('http://') or file_name.startswith('https://'):
        return True
    
    return False




from typing import List, Union,Optional

import pandas as pd

from mitosheet.types import MitoFrontendIndexAndSelections


def get_selected_element(dfs: List[pd.DataFrame], indexAndSelections: Optional[MitoFrontendIndexAndSelections]) -> Union[pd.DataFrame, pd.Series, None]:

    if indexAndSelections is None:
        return None

    selected_dataframe_index = indexAndSelections['selectedDataframeIndex']
    if selected_dataframe_index < 0 or selected_dataframe_index >= len(dfs):
        return None
    
    df = dfs[selected_dataframe_index]

    # If there are multiple selections, for now we only return the first one - for simplicity in return types
    selection = next(iter(indexAndSelections['selections']))

    # Selections have the format: {'startingRowIndex': 0, 'endingRowIndex': 0, 'startingColumnIndex': 5, 'endingColumnIndex': 5}

    smallerRowIndex = min(selection['startingRowIndex'], selection['endingRowIndex'])
    largerRowIndex = max(selection['startingRowIndex'], selection['endingRowIndex'])
    smallerColumnIndex = min(selection['startingColumnIndex'], selection['endingColumnIndex'])
    largerColumnIndex = max(selection['startingColumnIndex'], selection['endingColumnIndex'])

    # If the row indexes selected are both -1, we just return the column
    if smallerRowIndex == -1 and largerRowIndex == -1:
        return df.iloc[:, smallerColumnIndex:largerColumnIndex + 1]
    
    # If the column indexes selected are both -1, we just return the row
    if smallerColumnIndex == -1 and largerColumnIndex == -1:
        return df.iloc[smallerRowIndex:largerRowIndex + 1, :]
    
    # If one row index is -1, then we return the column
    if smallerRowIndex == -1:
        return df.iloc[:, smallerColumnIndex:largerColumnIndex + 1]
    
    # If one column index is -1, then we return the row
    if smallerColumnIndex == -1:
        return df.iloc[smallerRowIndex:largerRowIndex + 1, :]
    
    # Otherwise, we return the intersection of the row and column
    return df.iloc[smallerRowIndex:largerRowIndex + 1, smallerColumnIndex:largerColumnIndex + 1]



from datetime import datetime, timedelta
from typing import Any

import numpy as np
import pandas as pd

from mitosheet.errors import MitoError
from mitosheet.is_type_utils import (is_bool_dtype, is_datetime_dtype,
                                     is_float_dtype, is_int_dtype,
                                     is_string_dtype)
from mitosheet.public.v3.errors import handle_sheet_function_errors
from mitosheet.public.v3.sheet_functions.utils import \
    get_series_from_primitive_or_series
from mitosheet.public.v3.types.decorators import cast_values_in_arg_to_type
from mitosheet.public.v3.types.sheet_function_types import (
    AnyPrimitiveOrSeriesInputType, BoolRestrictedInputType,
    IntRestrictedInputType)


@handle_sheet_function_errors
def FILLNAN(series: pd.Series, replacement: AnyPrimitiveOrSeriesInputType) -> pd.Series:
    """
    {
        "function": "FILLNAN",
        "description": "Replaces the NaN values in the series with the replacement value.",
        "search_terms": ["fillnan", "nan", "fill nan", "missing values", "null", "null value", "fill null"],
        "examples": [
            "FILLNAN(A, 10)",
            "FILLNAN(A, 'replacement')"
        ],
        "syntax": "FILLNAN(series, replacement)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to replace the NaN values in."
            },
            {
                "element": "replacement",
                "description": "A string, number, or date to replace the NaNs with."
            }
        ]
    }
    """
    return series.fillna(replacement)


@handle_sheet_function_errors
def TYPE(series: pd.Series) -> pd.Series:
    """
    {
        "function": "TYPE",
        "description": "Returns the type of each element of the passed series. Return values are 'number', 'str', 'bool', 'datetime', 'object', or 'NaN'.",
        "search_terms": ["type", "dtype"],
        "examples": [
            "TYPE(Nums_and_Strings)",
            "IF(TYPE(Account_Numbers) != 'NaN', Account_Numbers, 0)"
        ],
        "syntax": "TYPE(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to get the type of each element of."
            }
        ]
    }
    """

    def get_element_type(element):
        # Start with bool!
        if isinstance(element, bool):
            return 'bool'
        elif isinstance(element, int):
            return 'number'
        elif isinstance(element, float):
            if np.isnan(element):
                return 'NaN'
            return 'number'
        elif element is None or element is pd.NaT:
            return 'NaN'
        elif isinstance(element, str):
            return 'string'
        elif isinstance(element, datetime) or isinstance(element, pd.Timestamp):
            return 'datetime'
        elif isinstance(element, timedelta) or isinstance(element, pd.Timedelta):
            return 'timedelta'
        return 'object'

    return series.apply(get_element_type).astype('str')



@handle_sheet_function_errors
def GETPREVIOUSVALUE(series: pd.Series, condition: BoolRestrictedInputType) -> pd.Series:
    """
    {
        "function": "GETPREVIOUSVALUE",
        "description": "Returns the value from series that meets the condition.",
        "search_terms": ["ffill"],
        "examples": [
            "GETPREVIOUSVALUE(Max_Balances, Max_Balances > 0)"
        ],
        "syntax": "GETPREVIOUSVALUE(series, condition)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to get the previous value from."
            }, {
                "element": "condition",
                "description": "When condition is True, a new previous value is set, and carried forward until the condition is True again."
            }
        ]
    }
    """
    # Default to a different last occurence depending on the type
    column_dtype = str(series.dtype)
    last_occurrence: Any = -1
    if is_int_dtype(column_dtype) or is_float_dtype(column_dtype):
        last_occurrence = -1
    elif is_string_dtype(column_dtype):
        last_occurrence = ''
    elif is_bool_dtype(column_dtype):
        last_occurrence = False
    elif is_datetime_dtype(column_dtype):
        last_occurrence = pd.NaT

    condition = get_series_from_primitive_or_series(condition, series.index)

    result = []
    for index, value in condition.items():
        if value:
            last_occurrence = series[index]
        result.append(last_occurrence)

    return pd.Series(result, index=series.index)

@handle_sheet_function_errors
def GETNEXTVALUE(series: pd.Series, condition: BoolRestrictedInputType) -> pd.Series:
    """
    {
        "function": "GETNEXTVALUE",
        "description": "Returns the next value from series that meets the condition.",
        "search_terms": ["ffill"],
        "examples": [
            "GETNEXTVALUE(Max_Balances, Max_Balances > 0)"
        ],
        "syntax": "GETNEXTVALUE(series, condition)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to get the next value from."
            }, {
                "element": "condition",
                "description": "When condition is True, a new previous value is set, and carried backwards until the condition is True again."
            }
        ]
    }
    """
    reversed_series = series[::-1]
    condition = get_series_from_primitive_or_series(condition, series.index)
    reversed_condition = condition[::-1]

    return GETPREVIOUSVALUE(reversed_series, reversed_condition)[::-1]

@cast_values_in_arg_to_type('index', 'int')
def VLOOKUP(lookup_value: AnyPrimitiveOrSeriesInputType, where: pd.DataFrame, index: IntRestrictedInputType) -> pd.Series:
    """
    {
        "function": "VLOOKUP",
        "description": "Looks up a value in a range and returns the value in the same row from a column you specify.",
        "search_terms": ["vlookup", "merge", "join", "search", "lookup"],
        "category": "REFERENCE",
        "examples": [
            "VLOOKUP(Names0, Ids:Ages, 1)",
            "VLOOKUP('John Smith', Names:Ages, 2)",
            "VLOOKUP(Names0, Ids:Ages, Column Indexes0)"
        ],
        "syntax": "VLOOKUP(lookup_value, where, index)",
        "syntax_elements": [{
                "element": "lookup_value",
                "description": "The value to look up."
            }, {
                "element": "where",
                "description": "The range to look up in."
            }, {
                "element": "index",
                "description": "The column index to return."
            }
        ]
    }
    """
    where_first_column_case_insensitive = where.iloc[:,0].copy()
    where_first_column_case_insensitive.name = str(where_first_column_case_insensitive.name) + 'MITO_CASE_INSENSITIVE'

    # If the lookup value and index are both a primitive, we don't need to merge. 
    if not isinstance(lookup_value, pd.Series) and isinstance(index, int):
        if type(lookup_value) != type(where.iloc[0,0]):
            raise MitoError(
                'invalid_args_error',
                'VLOOKUP',
                f'VLOOKUP requires the lookup value and the first column of the where range to be the same type. The lookup value is of type {type(lookup_value)} and the first column of the where range is of type {type(where.iloc[0,0])}.'
            )

        # If the lookup value and the first column are strings, make them lowecase for case-insensitive matching
        if isinstance(lookup_value, str) and isinstance(where.iloc[0,0], str):
            lookup_value = lookup_value.lower()
            where_first_column_case_insensitive = where_first_column_case_insensitive.str.lower()

        matching_row = where.loc[where_first_column_case_insensitive == lookup_value]
        if matching_row.empty:
            return None
        else:
            return matching_row.iloc[0, index-1]

    value = get_series_from_primitive_or_series(lookup_value, where.index)

    value.name = 'lookup_value'
    indices_to_return_from_range = get_series_from_primitive_or_series(index, value.index)

    # If the lookup value and the first column of the where range are different types, we raise an error
    if value.dtype != where_first_column_case_insensitive.dtype:
        raise MitoError(
            'invalid_args_error',
            'VLOOKUP',
            f'VLOOKUP requires the lookup value and the first column of the where range to be the same type. The lookup value is of type {value.dtype} and the first column of the where range is of type {where_first_column_case_insensitive.dtype}.'
        )

    # If the series is a string, convert it to lowercase because Excel's vlookup is case insensitive
    if is_string_dtype(str(value.dtype)):
        value = value.str.lower()
        where_first_column_case_insensitive = where_first_column_case_insensitive.str.lower()

    # Add where_first_column_case_insensitive to the front of the dataframe so we can use the case insensitive merge 
    # without effecting the return values
    where = pd.concat([where_first_column_case_insensitive, where], axis=1)
    indices_to_return_from_range = indices_to_return_from_range + 1

    where_deduplicated = where.drop_duplicates(subset=where_first_column_case_insensitive.name)
    
    # Update first column to use the deduplicated version
    where_first_column = where_deduplicated.iloc[:,0]
    
    # Then merge on the column we're looking up from, and the df we're looking up in.
    merged = pd.merge(value, where_deduplicated, left_on='lookup_value', right_on=where_first_column, how='left')

    # Change the indexes back to the indexes of the lookup value so the results 
    # can be added back to the calling dataframe.
    merged.index = value.index
    
    def get_value_at_index_in_row(row):
        try:
            return row.iloc[indices_to_return_from_range[row.name]]
        # Because we can't control what the user puts in the index, we need to catch any errors
        except Exception:
            return None
    return merged.apply(get_value_at_index_in_row, axis=1)

# TODO: we should see if we can list these automatically!
MISC_FUNCTIONS = {
    'FILLNAN': FILLNAN,
    'GETPREVIOUSVALUE': GETPREVIOUSVALUE,
    'GETNEXTVALUE': GETNEXTVALUE,
    'TYPE': TYPE,
    'VLOOKUP': VLOOKUP
}





from mitosheet.public.v3.sheet_functions.number_functions import *
from mitosheet.public.v3.sheet_functions.string_functions import *
from mitosheet.public.v3.sheet_functions.date_functions import *
from mitosheet.public.v3.sheet_functions.bool_functions import *
from mitosheet.public.v3.sheet_functions.misc_functions import *

FUNCTIONS = dict(NUMBER_FUNCTIONS, **STRING_FUNCTIONS, **DATE_FUNCTIONS, **CONTROL_FUNCTIONS, **MISC_FUNCTIONS)

# Overwrite __all__ so when you run from mitosheet.sheet_functions import *, it just imports the functions themselves!
__all__ = [
    func for func in FUNCTIONS.keys()
]


from dash import Dash, html, Input, Output, Patch, callback
from mitosheet.mito_dash.v1 import Spreadsheet, activate_mito

app = Dash(__name__)
activate_mito(app)

app.layout = html.Div([
    html.Button("Add Spreadsheet", id="dynamic-add-spreadsheet-btn", n_clicks=0),
    html.Div(id='dynamic-spreadsheet-container-div', children=[]),
])

@callback(
    Output('dynamic-spreadsheet-container-div', 'children'),
    Input('dynamic-add-spreadsheet-btn', 'n_clicks')
)
def display_dropdowns(n_clicks):
    patched_children = Patch()

    new_element = html.Div([
        Spreadsheet(
            id={
                'type': 'spreadsheet',
                'id': str(n_clicks)
            },
            import_folder='datasets'
        ),
    ])
    patched_children.append(new_element)
    return patched_children


if __name__ == '__main__':
    app.run(debug=True)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from copy import deepcopy
from distutils.version import LooseVersion
from typing import Any, Collection, Dict, List, Optional, Tuple

import pandas as pd

from mitosheet.array_utils import deduplicate_array
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.filter_code_chunk import (
    combine_filter_strings, get_single_filter_string)
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import (
    NEWLINE_TAB, get_column_header_list_as_transpiled_code,
    get_column_header_as_transpiled_code)
from mitosheet.types import (ColumnHeader, ColumnHeaderWithFilter,
                             ColumnHeaderWithPivotTransform, ColumnID,
                             ColumnIDWithFilter, ColumnIDWithPivotTransform)
from mitosheet.utils import is_prev_version

USE_INPLACE_PIVOT = tuple([int(i) for i in pd.__version__.split('.')]) < (1, 5, 0)


def get_new_column_header_from_column_header_with_pivot_transform(chwpt: ColumnHeaderWithPivotTransform) -> ColumnHeader:
    from mitosheet.step_performers.pivot import PCT_NO_OP

    column_header, transformation = chwpt['column_header'], chwpt['transformation']
    # We need to turn the column header into a string before creating the new one, so that we can
    # append to it for the new temporary transformation column
    if transformation == PCT_NO_OP:
        return column_header

    return f'{str(column_header)} ({transformation})'


def get_flatten_code(use_inplace_pivot: bool) -> str:
    # Helpful constants for code formatting. The in_place parameter was depricated
    # since 1.5.0, so we use a different method for formatting in this case. We keep
    # it for earlier versions, as the set_axis functioned required this on pre 1.0 to
    # avoid ending up with a dataframe with no axis
    if use_inplace_pivot:
        return f'pivot_table.set_axis([flatten_column_header(col) for col in pivot_table.keys()], axis=1, inplace=True)'
    else:
        return f'pivot_table = pivot_table.set_axis([flatten_column_header(col) for col in pivot_table.keys()], axis=1)'


def values_to_functions_code(values: Dict[ColumnHeader, Collection[str]]) -> str:
    """
    Helper function for turning the values mapping sent by the frontend to the values
    mapping that works in generated code. Namely, needs to replay Count Unique with the
    function pd.Series.nunique.
    """
    string_values = f'{values}'
    # NOTE: this needs to match the values sent from the frontend
    # also note that we overwrite the quotes around Count Unique
    return string_values.replace('\'count unique\'', 'pd.Series.nunique')

def build_args_code(
        pivot_rows_with_transforms: List[ColumnHeaderWithPivotTransform],
        pivot_columns_with_transforms: List[ColumnHeaderWithPivotTransform],
        values: Dict[ColumnHeader, Collection[str]]
    ) -> str:
    """
    Helper function for building an arg string, while leaving
    out empty arguments. 
    """

    # Because there might have been temporary columns created by the pivot
    # transformations, we need to use these in our final args usage. NOTE: as in 
    # execution, we have to deduplicate
    final_pivot_rows = deduplicate_array([get_new_column_header_from_column_header_with_pivot_transform(chwpt) for chwpt in pivot_rows_with_transforms])
    final_pivot_columns = deduplicate_array([get_new_column_header_from_column_header_with_pivot_transform(chwpt) for chwpt in pivot_columns_with_transforms])

    args = []
    if len(final_pivot_rows) > 0:
        args.append(f'index={get_column_header_list_as_transpiled_code(final_pivot_rows)},')

    if len(final_pivot_columns) > 0:
        args.append(f'columns={get_column_header_list_as_transpiled_code(final_pivot_columns)},')

    if len(values) > 0:
        values_keys = list(values.keys())
        args.append(f'values={get_column_header_list_as_transpiled_code(values_keys)},')
        args.append(f'aggfunc={values_to_functions_code(values)}')
        
    return NEWLINE_TAB.join(args)


class PivotCodeChunk(CodeChunk):

    def __init__(
        self, 
        prev_state: State, 
        sheet_index: int,
        destination_sheet_index: Optional[int],
        pivot_rows_column_ids_with_transforms: List[ColumnIDWithPivotTransform],
        pivot_columns_column_ids_with_transforms: List[ColumnIDWithPivotTransform],
        pivot_filters: List[ColumnIDWithFilter],
        values_column_ids_map: Dict[ColumnID, Collection[str]],
        flatten_column_headers: Optional[bool],
        public_interface_version: int,
        new_df_name: str,
        optional_code_that_successfully_executed: Optional[Tuple[List[str], List[str]]] = None
    ):
        super().__init__(prev_state)
        self.sheet_index = sheet_index
        self.destination_sheet_index = destination_sheet_index
        self.pivot_rows_column_ids_with_transforms = pivot_rows_column_ids_with_transforms
        self.pivot_columns_column_ids_with_transforms = pivot_columns_column_ids_with_transforms
        self.pivot_filters_ids = pivot_filters
        self.values_column_ids_map = values_column_ids_map
        self.flatten_column_headers = flatten_column_headers
        self.public_interface_version = public_interface_version

        if optional_code_that_successfully_executed is not None:
            self.optional_code_that_successfully_executed = optional_code_that_successfully_executed

        # On earlier pandas versions (e.g. 0.24.2), the pivot table function returned
        # a series from the above function call. Thus, we need to move it to a df for
        # all our other code run properly on it. This code should only run in early 
        # versions of pandas
        self.was_series = LooseVersion(pd.__version__) <= LooseVersion('1.0.0')

        self.old_df_name = self.prev_state.df_names[self.sheet_index]
        self.new_df_name = new_df_name

    def get_display_name(self) -> str:
        return 'Pivoted'
    
    def get_description_comment(self) -> str:
        return f'Pivoted {self.old_df_name} into {self.new_df_name}'

    def get_code(self) -> Tuple[List[str], List[str]]:
    
        # Get just the column headers in a list, for convenience
        pivot_rows = [self.prev_state.column_ids.get_column_header_by_id(self.sheet_index, cit['column_id']) for cit in self.pivot_rows_column_ids_with_transforms]
        pivot_columns = [self.prev_state.column_ids.get_column_header_by_id(self.sheet_index, cit['column_id']) for cit in self.pivot_columns_column_ids_with_transforms]

        # Make new objects with all columns headers
        pivot_rows_with_transforms: List[ColumnHeaderWithPivotTransform] = [{
            'column_header': self.prev_state.column_ids.get_column_header_by_id(self.sheet_index, chwpt['column_id']),
            'transformation': chwpt['transformation']
        } for chwpt in self.pivot_rows_column_ids_with_transforms]
        pivot_columns_with_transforms: List[ColumnHeaderWithPivotTransform] = [{
            'column_header': self.prev_state.column_ids.get_column_header_by_id(self.sheet_index, chwpt['column_id']),
            'transformation': chwpt['transformation']
        } for chwpt in self.pivot_columns_column_ids_with_transforms]
        values = {
            self.prev_state.column_ids.get_column_header_by_id(self.sheet_index, column_id): value 
            for column_id, value in self.values_column_ids_map.items()
        }
        pivot_filters: List[ColumnHeaderWithFilter] = [
            {
                'column_header': self.prev_state.column_ids.get_column_header_by_id(self.sheet_index, pf['column_id']), 
                'column_dtype': str(self.prev_state.dfs[self.sheet_index][pf['column_id']].dtype),
                'filter': pf['filter']
            }
            for pf in self.pivot_filters_ids
        ]

        # If there are no keys or values to aggregate on we return an empty dataframe. 
        if len(pivot_rows_with_transforms) == 0 and len(pivot_columns_with_transforms) == 0 or len(values) == 0:
            return [f'{self.new_df_name} = pd.DataFrame(data={{}})'], ['import pandas as pd']

        transpiled_code = []

        # First, filter down to the rows of the original dataframe that we need
        if len(pivot_filters) > 0:
            filter_strings = [
                get_single_filter_string(self.old_df_name, pf['column_header'], pf['column_dtype'], pf['filter'])
                for pf in pivot_filters
            ]
            full_filter_string = combine_filter_strings('And', filter_strings)
            transpiled_code.append(f'tmp_df = {self.old_df_name}[{full_filter_string}]')
            old_df_name = 'tmp_df' # update the old_df name, for the step below
        else:
            old_df_name = self.old_df_name

        # Drop any columns we don't need, to avoid issues where pandas freaks out
        # and says there is a non-1-dimensional grouper
        column_headers_list = get_column_header_list_as_transpiled_code(list(set(pivot_rows + pivot_columns + list(values.keys()))))
        transpiled_code.append(f'tmp_df = {old_df_name}[{column_headers_list}].copy()')

        # Create any new temporary columns that are formed by the pivot transforms
        transpiled_code = transpiled_code + get_code_for_transform_columns('tmp_df', pivot_rows_with_transforms) + get_code_for_transform_columns('tmp_df', pivot_columns_with_transforms)

        # Do the actual pivot
        pivot_table_args = build_args_code(pivot_rows_with_transforms, pivot_columns_with_transforms, values)
        transpiled_code.append(f'pivot_table = tmp_df.pivot_table({NEWLINE_TAB}{pivot_table_args}\n)')

        if self.was_series:
            # TODO: do we want a comment to explain this?
            transpiled_code.append(f'pivot_table = pd.DataFrame(pivot_table)')

        if self.flatten_column_headers:
            # Flatten column headers, which we always do because it's hard to tell when we should
            transpiled_code.append(get_flatten_code(USE_INPLACE_PIVOT))

        # Finially, reset the column name, and the indexes!
        transpiled_code.append(f'{self.new_df_name} = pivot_table.reset_index()')

        return transpiled_code, [] # TODO: we might actually need pd to be defined!

    def _combine_right_with_pivot_code_chunk(self, pivot_code_chunk: "PivotCodeChunk") -> Optional["CodeChunk"]:
        """
        We can combine a pivot code chunk with the one before it if the destination
        sheet index of the is the created code index of this step.
        """
        destination_sheet_index = self.destination_sheet_index
        other_destination_sheet_index = pivot_code_chunk.destination_sheet_index

        # If both of the pivots are overwriting the same destination sheet index, and they are both defined
        if destination_sheet_index is not None and destination_sheet_index == other_destination_sheet_index:
            return PivotCodeChunk(
                self.prev_state,
                pivot_code_chunk.sheet_index,
                pivot_code_chunk.destination_sheet_index,
                pivot_code_chunk.pivot_rows_column_ids_with_transforms,
                pivot_code_chunk.pivot_columns_column_ids_with_transforms,
                pivot_code_chunk.pivot_filters_ids,
                pivot_code_chunk.values_column_ids_map,
                pivot_code_chunk.flatten_column_headers,
                pivot_code_chunk.public_interface_version,
                pivot_code_chunk.new_df_name,
                pivot_code_chunk.optional_code_that_successfully_executed
            )

        # If one of the pivots if creating the code chunk that the new one is overwriting, then we can optimize
        # this as well
        created_sheet_index = self.get_created_sheet_indexes()
        if created_sheet_index is not None and created_sheet_index[0] == other_destination_sheet_index:
            return PivotCodeChunk(
                self.prev_state,
                pivot_code_chunk.sheet_index,
                pivot_code_chunk.destination_sheet_index,
                pivot_code_chunk.pivot_rows_column_ids_with_transforms,
                pivot_code_chunk.pivot_columns_column_ids_with_transforms,
                pivot_code_chunk.pivot_filters_ids,
                pivot_code_chunk.values_column_ids_map,
                pivot_code_chunk.flatten_column_headers,
                pivot_code_chunk.public_interface_version,
                pivot_code_chunk.new_df_name,
                pivot_code_chunk.optional_code_that_successfully_executed
            )

        return None

    def combine_right(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        if isinstance(other_code_chunk, PivotCodeChunk):
            return self._combine_right_with_pivot_code_chunk(other_code_chunk)
        return None

    def combine_left(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        # Because overwriting a pivot overwrites all the edits on that pivot table
        # we can optimize out any edits that are before the pivot 
        # NOTE: if we start carrying edits on pivots forward, we should remove this 
        # optimization

        destination_sheet_index = self.destination_sheet_index
        edited_sheet_indexes = other_code_chunk.get_edited_sheet_indexes()

        if edited_sheet_indexes is not None and len(edited_sheet_indexes) == 1 and edited_sheet_indexes[0] == destination_sheet_index:
            return PivotCodeChunk(
                other_code_chunk.prev_state,
                self.sheet_index,
                self.destination_sheet_index,
                self.pivot_rows_column_ids_with_transforms,
                self.pivot_columns_column_ids_with_transforms,
                self.pivot_filters_ids,
                self.values_column_ids_map,
                self.flatten_column_headers,
                self.public_interface_version,
                self.new_df_name,
                self.optional_code_that_successfully_executed
            )

        return None

    def get_created_sheet_indexes(self) -> Optional[List[int]]:
        if self.destination_sheet_index is None:
            return [len(self.prev_state.dfs)]
        else:
            # Note: editing a dataframe does not create a sheet index, it 
            # overwrites it instead. See get_edited_sheet_indexes below
            return None
        
    def get_source_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]

    def get_edited_sheet_indexes(self) -> List[int]:
        if self.destination_sheet_index is not None:
            return [self.destination_sheet_index]
        return []
    
    def can_be_reordered_with(self, code_chunk: CodeChunk) -> bool:
        """
        A pivot either creates or edits a sheet. In either case, we don't
        want to reorder around it as this may cause bugs -- so we don't. 
        """
        return False


def get_code_for_transform_columns(df_name: str, column_headers_with_transforms: List[ColumnHeaderWithPivotTransform]) -> List[str]:
    from mitosheet.step_performers.pivot import (
        PCT_DATE_DAY_HOUR, PCT_DATE_DAY_OF_MONTH, PCT_DATE_DAY_OF_WEEK,
        PCT_DATE_HOUR, PCT_DATE_HOUR_MINUTE, PCT_DATE_MINUTE, PCT_DATE_MONTH,
        PCT_DATE_MONTH_DAY, PCT_DATE_QUARTER, PCT_DATE_SECOND, PCT_DATE_WEEK,
        PCT_DATE_YEAR, PCT_DATE_YEAR_MONTH, PCT_DATE_YEAR_MONTH_DAY,
        PCT_DATE_YEAR_MONTH_DAY_HOUR, PCT_DATE_YEAR_MONTH_DAY_HOUR_MINUTE,
        PCT_DATE_YEAR_QUARTER, PCT_NO_OP)        

    code = []
    for chwpt in column_headers_with_transforms:
        column_header, transformation = chwpt['column_header'], chwpt['transformation']
        if transformation == PCT_NO_OP:
            continue

        # We need to turn the column header into a string before creating the new one, so that we can
        # append to it for the new temporary transformation column
        new_column_header = get_new_column_header_from_column_header_with_pivot_transform(chwpt)
        if transformation == PCT_DATE_YEAR:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.year')
        if transformation == PCT_DATE_QUARTER:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.quarter')
        if transformation == PCT_DATE_MONTH:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.month')
        if transformation == PCT_DATE_WEEK:
            if is_prev_version(pd.__version__, '1.0.0'):
                code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.week')
            else:
                code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.isocalendar().week.astype(int)')
        if transformation == PCT_DATE_DAY_OF_MONTH:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.day')
        if transformation == PCT_DATE_DAY_OF_WEEK:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.weekday')
        if transformation == PCT_DATE_HOUR:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.hour')
        if transformation == PCT_DATE_MINUTE:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.minute')
        if transformation == PCT_DATE_SECOND:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.second')
        if transformation == PCT_DATE_YEAR_MONTH_DAY_HOUR_MINUTE:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.strftime("%Y-%m-%d %H:%M")')
        if transformation == PCT_DATE_YEAR_MONTH_DAY_HOUR:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.strftime("%Y-%m-%d %H")')
        if transformation == PCT_DATE_YEAR_MONTH_DAY:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.strftime("%Y-%m-%d")')
        if transformation == PCT_DATE_YEAR_MONTH:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.strftime("%Y-%m")')
        if transformation == PCT_DATE_YEAR_QUARTER:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.year.astype(str) + "-Q" + {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.quarter.astype(str)')
        if transformation == PCT_DATE_MONTH_DAY:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.strftime("%m-%d")')
        if transformation == PCT_DATE_DAY_HOUR:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.strftime("%d %H")')
        if transformation == PCT_DATE_HOUR_MINUTE:
            code.append(f'{df_name}[{get_column_header_as_transpiled_code(new_column_header)}] = {df_name}[{get_column_header_as_transpiled_code(column_header)}].dt.strftime("%H:%M")')

    return code



import json
import pytest
import pandas as pd
from mitosheet.enterprise.mito_config import MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH, MITO_CONFIG_VERSION
from mitosheet.saved_analyses.save_utils import write_save_analysis_file
from mitosheet.tests.test_mito_config import delete_all_mito_config_environment_variables

from mitosheet.tests.test_utils import create_mito_wrapper


def test_can_pass_sheet_function():
    def ADD1(col):
        return col + 1 
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}), sheet_functions=[ADD1])

    mito.set_formula('=ADD1(A0)', 0, 'B', add_column=True)
    assert mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [2, 3, 4]}))

def test_can_pass_multiple_sheet_functions():
    def ADD1(col):
        return col + 1 
    def ADD2(col):
        return col + 2
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}), sheet_functions=[ADD1, ADD2])

    mito.set_formula('=ADD2(ADD1(A0))', 0, 'B', add_column=True)
    assert mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6]}))


def test_pass_sheet_function_then_replay_errors_if_not_passed_again():
    def ADD1(col):
        return col + 1 
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, sheet_functions=[ADD1])

    mito.set_formula('=ADD1(A0)', 0, 'B', add_column=True)
    assert mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [2, 3, 4]}))

    analysis_name = mito.mito_backend.analysis_name
    write_save_analysis_file(mito.mito_backend.steps_manager)

    new_mito = create_mito_wrapper(df)
    new_mito.replay_analysis(analysis_name)

    assert new_mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3]}))

def test_pass_sheet_function_then_replay_works_if_passed_again():
    def ADD1(col):
        return col + 1 
    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, sheet_functions=[ADD1])

    mito.set_formula('=ADD1(A0)', 0, 'B', add_column=True)
    assert mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [2, 3, 4]}))

    analysis_name = mito.mito_backend.analysis_name
    write_save_analysis_file(mito.mito_backend.steps_manager)

    new_mito = create_mito_wrapper(df, sheet_functions=[ADD1])
    new_mito.replay_analysis(analysis_name)

    assert new_mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [2, 3, 4]}))

def test_sheet_functions_with_non_caps_error():
    def lower(col):
        return col + 1 
    df = pd.DataFrame({'A': [1, 2, 3]})
    with pytest.raises(ValueError):
        mito = create_mito_wrapper(df, sheet_functions=[lower])


def test_pass_sheet_function_returns_list():
    def COLUMN_SUM(df):
        return [df.sum() for x in range(len(df))]

    df = pd.DataFrame({'A': [1, 2, 3]})
    mito = create_mito_wrapper(df, sheet_functions=[COLUMN_SUM])

    mito.set_formula('=COLUMN_SUM(A0)', 0, 'B', add_column=True)
    assert mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [6, 6, 6]}))


def test_user_defined_function_valid_doc_string():
    def ADD1(col):
        """Adds 1 to the column"""
        return col + 1 
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}), sheet_functions=[ADD1])
    documentation = json.loads(mito.analysis_data_json)['userDefinedFunctions']

    assert documentation == [{'function': 'ADD1', 'description': 'Adds 1 to the column', 'search_terms': ['Adds', 'the', 'column'], 'syntax': 'ADD1(col)', 'syntax_elements': [{'element': 'col', 'description': ''}]}]

def test_user_defined_function_function_documentation_object():
    def ADD1(col):
        """
        {
            "function": "ADD1",
            "description": "Returns the absolute value of the passed number or series.",
            "search_terms": ["abs", "absolute value"],
            "examples": [
                "ABS(-1.3)",
                "ABS(A)"
            ],
            "syntax": "ABS(value)",
            "syntax_elements": [{
                    "element": "value",
                    "description": "The value or series to take the absolute value of."
                }
            ]
        }
        """
        return col + 1 
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}), sheet_functions=[ADD1])
    documentation = json.loads(mito.analysis_data_json)['userDefinedFunctions']

    assert documentation == [{'function': 'ADD1', 'description': 'Returns the absolute value of the passed number or series.', 'search_terms': ['abs', 'absolute value'], 'examples': ['ABS(-1.3)', 'ABS(A)'], 'syntax': 'ABS(value)', 'syntax_elements': [{'element': 'value', 'description': 'The value or series to take the absolute value of.'}]}]

def test_user_defined_function_from_environment_variable():
    file_one = """
def ADD1(col):
    return col + 1
"""

    # Write these files to the folder
    with open('file_one.py', 'w') as f:
        f.write(file_one)

    # Set the environment variable
    import os
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH] = "./file_one.py"

    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}))
    mito.set_formula('=ADD1(A0)', 0, 'B', add_column=True)
    assert mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [2, 3, 4]}))

    # Reset the environmnet
    delete_all_mito_config_environment_variables()
    os.remove('file_one.py')


def test_user_defined_function_from_environment_variable_with_imports():
    file_one = """
def HELPER_FUNCTION():
    print(1)
"""

    file_two = """
from file_one import HELPER_FUNCTION

def ADD1(col):
    HELPER_FUNCTION()
    return col + 1

def ADD2(col):
    return ADD1(col) + 1
"""

    # Write these files to the folder
    with open('file_one.py', 'w') as f:
        f.write(file_one)
    with open('file_two.py', 'w') as f:
        f.write(file_two)

    # Set the environment variable
    import os
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH] = "./file_two.py"

    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3]}))
    mito.set_formula('=ADD2(A0)', 0, 'B', add_column=True)
    assert mito.dfs[0].equals(pd.DataFrame({'A': [1, 2, 3], 'B': [3, 4, 5]}))

    assert len(json.loads(mito.mito_backend.steps_manager.analysis_data_json)["userDefinedFunctions"]) == 2

    # Reset the environmnet
    delete_all_mito_config_environment_variables()
    os.remove('file_one.py')
    os.remove('file_two.py')

def test_user_defined_function_from_environment_variable_and_mitosheet_call():
    file_one = """
def ADD1(col):
    return col + 1
"""

    # Write these files to the folder
    with open('file_one.py', 'w') as f:
        f.write(file_one)

    # Set the environment variable
    import os
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH] = "./file_one.py"

    def ADD2(col):
        """
        {
            "function": "ADD1",
            "description": "Returns the absolute value of the passed number or series.",
            "search_terms": ["abs", "absolute value"],
            "examples": [
                "ABS(-1.3)",
                "ABS(A)"
            ],
            "syntax": "ABS(value)",
            "syntax_elements": [{
                    "element": "value",
                    "description": "The value or series to take the absolute value of."
                }
            ]
        }
        """
        return col + 2
    mito = create_mito_wrapper(pd.DataFrame({'A': [1, 2, 3], 'B': [1, 2, 3]}), sheet_functions=[ADD2])
    mito.set_formula('=ADD1(A0)', 0, 'A')
    mito.set_formula('=ADD2(B0)', 0, 'B')

    assert mito.dfs[0].equals(pd.DataFrame({'A': [2, 3, 4], 'B': [3, 4, 5]}))

    # Reset the environmnet
    delete_all_mito_config_environment_variables()
    os.remove('file_one.py')




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.column_steps.change_column_dtype_code_chunk import \
    ChangeColumnDtypeCodeChunk
from mitosheet.is_type_utils import is_datetime_dtype, is_string_dtype
from mitosheet.public.v1.sheet_functions.types.utils import \
    get_to_datetime_params
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import ColumnID, StepType


class ChangeColumnDtypeStepPerformer(StepPerformer):
    """"
    A step that allows changing the dtype of a column to a different
    dtype.

    Currently, supports: 'bool', 'int', 'float', 'str', 'datetime', 'timedelta'
    """

    @classmethod
    def step_version(cls) -> int:
        return 4

    @classmethod
    def step_type(cls) -> str:
        return 'change_column_dtype'

    @classmethod
    def saturate(cls, prev_state: State, params: Dict[str, Any], previous_steps: List[StepType]) -> Dict[str, Any]:
        sheet_index: int = params['sheet_index']
        column_ids: List[ColumnID] = params['column_ids']

        # Save all the old dtypes
        old_dtypes = dict()
        for column_id in column_ids:
            column_header = prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)
            old_dtypes[column_id] = str(prev_state.dfs[sheet_index][column_header].dtype)

        params['old_dtypes'] = old_dtypes
        
        return params

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        column_ids: List[ColumnID] = get_param(params, 'column_ids')
        old_dtypes: Dict[ColumnID, str] = get_param(params, 'old_dtypes')
        new_dtype: str = get_param(params, 'new_dtype')
        public_interface_version: str = get_param(params, 'public_interface_version')

        to_datetime_params_map: Dict[ColumnID, Dict[str, Any]] = {}
        changed_column_ids = []
        for column_id in column_ids:
            column_header = prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)
            column = prev_state.dfs[sheet_index][column_header]
            column_dtype = str(column.dtype)

            if column_dtype != new_dtype:
                changed_column_ids.append(column_id)

                if is_string_dtype(column_dtype) and is_datetime_dtype(new_dtype):
                    if not is_datetime_dtype(column_dtype):
                        to_datetime_params = get_to_datetime_params(column)
                        to_datetime_params_map[column_id] = to_datetime_params

        execution_data: Dict[str, Any] = {
            'changed_column_ids': changed_column_ids
        }

        if len(to_datetime_params_map) > 0:
            execution_data['to_datetime_params_map'] = to_datetime_params_map

        return cls.execute_through_transpile(
            prev_state,
            params,
            execution_data
        )        

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ChangeColumnDtypeCodeChunk(
                prev_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'column_ids'),
                get_param(params, 'old_dtypes'),
                get_param(params, 'new_dtype'),
                get_param(execution_data if execution_data is not None else {}, 'changed_column_ids'),
                execution_data.get('to_datetime_params_map', None) if execution_data is not None else None,
                get_param(params, 'public_interface_version')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}




import json
from typing import Any, Dict, Optional

from mitosheet.types import StepsManagerType


def get_params(params: Dict[str, Any], steps_manager: StepsManagerType) -> Optional[Dict[str, Any]]:
    """
    Loops back over the steps, looking for a matching step. Specifically, it tries
    to find the most recent step with the given params passed.

    If execution_data_to_match is passed, then these keys are propagated into the
    params as well (e.g. for use in Pivot). This is a bit of a hack, but ok for now.
    """
    step_type = params['step_type']
    step_id_to_match = params['step_id_to_match']
    execution_data_to_match = params['execution_data_to_match']
    params_to_match = params['params_to_match'] if 'params_to_match' in params else None

    # Loop over the steps backwards, so that we get the most recent one
    found_params: Optional[Dict[str, Any]] = None
    for step in steps_manager.steps_including_skipped[:steps_manager.curr_step_idx + 1][::-1]:
        if step.step_type != step_type:
            continue

        if step.step_id == step_id_to_match:
            found_params = step.params
        
        if step.params is not None and params_to_match is not None and len(params_to_match) > 0:
            all_matched = True
            for key, value in params_to_match.items():
                if step.params.get(key) != value:
                    all_matched = False
            if all_matched:
                found_params = dict(**step.params)

        if step.execution_data is not None and len(execution_data_to_match) > 0:
            all_matched = True
            for key, value in execution_data_to_match.items():
                if step.execution_data.get(key) != value:
                    all_matched = False
                
            if all_matched:
                found_params = dict(**step.params)
                for key, value in execution_data_to_match.items():
                    found_params[key] = value


        if found_params is not None:
            return found_params
    
    # Return nothing, if there is no step that meets this criteria
    return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Increments the render_count, which is the number of times that a sheet
has been rendered for this specific analysis.

Note that this count persists for the lifetime of the backend object, 
meaning that we have the following conditions:

Increments when:
1. You make the first mitosheet.sheet call, causing a render.
2. The page is refreshed

Resets (because it's an entirely new backend) when:
1. The kernel is restarted for any reason and then you make a new mitosheet.sheet() call
"""

from typing import List
from mitosheet.types import StepsManagerType

RENDER_COUNT_UPDATE_EVENT = 'render_count_update'
RENDER_COUNT_UPDATE_PARAMS: List[str] = []

def execute_render_count_update(steps_manager: StepsManagerType) -> None:
    steps_manager.render_count += 1

RENDER_COUNT_UPDATE = {
    'event_type': RENDER_COUNT_UPDATE_EVENT,
    'params': RENDER_COUNT_UPDATE_PARAMS,
    'execute': execute_render_count_update
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List


def upgrade_excel_range_import_1_to_2(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]: 
    """
    Adds an end condition to 'upper left corner value' excel range imports, that defaults to
    end_condition: 'first_empty_cell'.


    OLD: 
    {
        "step_version": 1, 
        "step_type": "excel_range_import", 
        "params": {
            "file_path": str,
            "sheet_name": str,
            "range_imports": {
                "type": 'range' | 'upper left corner value',
                "df_name": str,
                "value": str
            }[]
        }
    }

    NEW: 
    {
        "step_version": 2, 
        "step_type": "excel_range_import", 
        "params": {
            "file_path": str,
            "sheet_name": str,
            "range_imports": {
                "type": 'range',
                "df_name": str,
                "value": str
            } | {
                "type": 'upper left corner value',
                "df_name": str,
                "value": str,
                "end_condition": {'type': 'first empty cell'} | {'type': 'bottom left corner value', 'value': Any}
            }[]
        }
    }
    """

    params = step['params']
    new_range_imports = []
    for range_import in params['range_imports']:
        if range_import['type'] == 'range':
            new_range_imports.append(range_import)
        else:
            range_import['end_condition'] = {'type': 'first empty cell'}
            new_range_imports.append(range_import)

    return [{
        "step_version": 2, 
        "step_type": "excel_range_import", 
        "params": params
    }] + later_steps

def upgrade_excel_range_import_2_to_3(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]: 
    """
    Adds an end condition to 'upper left corner value' excel range imports, that defaults to
    end_condition: 'first_empty_cell'.


    OLD: 
    {
        "step_version": 2, 
        "step_type": "excel_range_import", 
        "params": {
            "file_path": str,
            "sheet_name": str,
            "range_imports": {
                "type": 'range',
                "df_name": str,
                "value": str
            } | {
                "type": 'upper left corner value',
                "df_name": str,
                "value": str,
                "end_condition": {'type': 'first empty cell'} | {'type': 'bottom left corner value', 'value': Any},
            }[]
        }
    }

    NEW: 
    {
        "step_version": 3, 
        "step_type": "excel_range_import", 
        "params": {
            "file_path": str,
            "sheet_name": str,
            "range_imports": {
                "type": 'range',
                "df_name": str,
                "value": str
            } | {
                "type": 'upper left corner value',
                "df_name": str,
                "value": str,
                "end_condition": {'type': 'first empty cell'} | {'type': 'bottom left corner value', 'value': Any},
                'column_end_condition': {'type': 'first empty cell'} | {'type': 'num columns', 'value': int}
            }[]
        }
    }
    """

    params = step['params']
    new_range_imports = []
    for range_import in params['range_imports']:
        if range_import['type'] == 'range':
            new_range_imports.append(range_import)
        else:
            range_import['column_end_condition'] = {'type': 'first empty cell'}
            new_range_imports.append(range_import)

    return [{
        "step_version": 3, 
        "step_type": "excel_range_import", 
        "params": params
    }] + later_steps

def upgrade_excel_range_import_3_to_4(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]: 
    """
    Changes upper left corner value imports to dynamic imports, for better and more clear grouping.

    OLD:
    {
        "step_version": 3, 
        "step_type": "excel_range_import", 
        "params": {
            "file_path": str,
            "sheet_name": str,
            "range_imports": {
                "type": 'range',
                "df_name": str,
                "value": str
            } | {
                "type": 'upper left corner value',
                "df_name": str,
                "value": str,
                "end_condition": {'type': 'first empty cell'} | {'type': 'bottom left corner value', 'value': Any},
                'column_end_condition': {'type': 'first empty cell'} | {'type': 'num columns', 'value': int}
            }[]
        }
    }

    NEW:
    {
        "step_version": 4, 
        "step_type": "excel_range_import", 
        "params": {
            "file_path": str,
            "sheet_name": str,
            "range_imports": {
                "type": 'range',
                "df_name": str,
                "value": str
            } | {
                "type": 'dynamic',
                "df_name": str,
                "start_condition": {
                    "type": 'upper left corner value',
                    "value": str,
                }
                "end_condition": {'type': 'first empty cell'} | {'type': 'bottom left corner value', 'value': Any},
                'column_end_condition': {'type': 'first empty cell'} | {'type': 'num columns', 'value': int}
            }[]
        }
    }
    """

    params = step['params']
    new_range_imports = []
    for range_import in params['range_imports']:
        if range_import['type'] == 'range':
            new_range_imports.append(range_import)
        else:
            range_import['type'] = 'dynamic'
            range_import['start_condition'] = {
                'type': 'upper left corner value',
                'value': range_import['value']
            }
            del range_import['value']
            new_range_imports.append(range_import)

    return [{
        "step_version": 4, 
        "step_type": "excel_range_import", 
        "params": params
    }] + later_steps

def upgrade_excel_range_import_4_to_5(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]: 
    """
    Adds the parameter convert_csv_to_xlsx and defaults it to false

    OLD:
    {
        "step_version": 4, 
        "step_type": "excel_range_import", 
        "params": {
            "file_path": str,
            "sheet_name": str,
            "range_imports": {
                "type": 'range',
                "df_name": str,
                "value": str
            } | {
                "type": 'dynamic',
                "df_name": str,
                "start_condition": {
                    "type": 'upper left corner value',
                    "value": str,
                }
                "end_condition": {'type': 'first empty cell'} | {'type': 'bottom left corner value', 'value': Any},
                'column_end_condition': {'type': 'first empty cell'} | {'type': 'num columns', 'value': int}
            }[]
        }
    }

    NEW:
    {
        "step_version": 5, 
        "step_type": "excel_range_import", 
        "params": {
            "file_path": str,
            "sheet_name": str,
            "convert_csv_to_xlsx": bool,
            "range_imports": {
                "type": 'range',
                "df_name": str,
                "value": str
            } | {
                "type": 'dynamic',
                "df_name": str,
                "start_condition": {
                    "type": 'upper left corner value',
                    "value": str,
                }
                "end_condition": {'type': 'first empty cell'} | {'type': 'bottom left corner value', 'value': Any},
                'column_end_condition': {'type': 'first empty cell'} | {'type': 'num columns', 'value': int}
            }[]
        }
    }
    """

    params = step['params']
    params['convert_csv_to_xlsx'] = False

    return [{
        "step_version": 5, 
        "step_type": "excel_range_import", 
        "params": params
    }] + later_steps

def upgrade_excel_range_import_5_to_6(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]: 
    """
    Changes the sheet name to be a type either of a sheet name or of a sheet index.

    OLD:
    {
        "step_version": 5, 
        "step_type": "excel_range_import", 
        "params": {
            "file_path": str,
            "sheet_name": str,
            "convert_csv_to_xlsx": bool,
            "range_imports": {
                "type": 'range',
                "df_name": str,
                "value": str
            } | {
                "type": 'dynamic',
                "df_name": str,
                "start_condition": {
                    "type": 'upper left corner value',
                    "value": str,
                }
                "end_condition": {'type': 'first empty cell'} | {'type': 'bottom left corner value', 'value': Any},
                'column_end_condition': {'type': 'first empty cell'} | {'type': 'num columns', 'value': int}
            }[]
        }
    }

    NEW:
    {
        "step_version": 6, 
        "step_type": "excel_range_import", 
        "params": {
            "file_path": str,
            "sheet": {
                "type": 'sheet name' | 'sheet index',
                "value": str | int
            }
            "convert_csv_to_xlsx": bool,
            "range_imports": {
                "type": 'range',
                "df_name": str,
                "value": str
            } | {
                "type": 'dynamic',
                "df_name": str,
                "start_condition": {
                    "type": 'upper left corner value',
                    "value": str,
                }
                "end_condition": {'type': 'first empty cell'} | {'type': 'bottom left corner value', 'value': Any},
                'column_end_condition': {'type': 'first empty cell'} | {'type': 'num columns', 'value': int}
            }[]
        }
    }
    """

    params = step['params']
    params['sheet'] = {
        'type': 'sheet name',
        'value': params['sheet_name']
    }
    del params['sheet_name']

    return [{
        "step_version": 6, 
        "step_type": "excel_range_import", 
        "params": params
    }] + later_steps




from datetime import datetime, timedelta
from typing import Callable, Union

import pandas as pd


class RollingRange():
    """
    A rolling range is a helper object that is passed to sheet functions when the user
    references a range. 

    Some examples:
    - A0 = SUM(B0:B0) => SUM(RollingRange(df[['B']], 1, 0))
    - A0 = SUM(B0:B1) => SUM(RollingRange(df[['B']], 2, 0))
    - A1 = SUM(B0:B1) => SUM(RollingRange(df[['B']], 2, -1))
    - A1 = SUM(B0:B2) => SUM(RollingRange(df[['B']], 3, -1))

    We use this object rather than pandas .shifts and rolling windows for a simple reason. Consider
    the formula A1 = SUM(B0:B2). 

    In this case:
    1.  if we shift df[['B']] before calling .rolling, we will loose some of the values 
    in the column B - but this formula clearly uses all of them. 
    2.  if we call .rolling on df[['B']] first, then we loose the ability to shift the 
        object in a single expression. But it needs to be in a single expression as 
        it's inside a SUM function.

    Thus, we need to attach both the window size and the offset in a single expression, and
    thus this is exactly what we capture in this new object.
    """

    def __init__(self, obj: pd.DataFrame, window: int, offset: int):
        self.obj = obj
        self.window = window
        self.offset = offset

    def apply(self, func: Callable[[pd.DataFrame], Union[str, float, int, bool, datetime, timedelta]], default_value: Union[str, float, int, bool, datetime, timedelta]=0) -> pd.Series:
        """
        Calls the func with each of the windows, and returns a series with
        the same index as the original dataframe.
        """
        result = []

        # Then, we get each window, and call the function with it
        start = 0 + self.offset
        end = start + self.window

        while (start - self.offset) < len(self.obj): 
            df_subset = self.obj[max(0, start):end] # avoid negative start, as this wraps around to the end

            # We manually detect the default value case, as it messes up types otherwise (e.g. .sum().sum() returns a float with an empty df)
            if len(df_subset) == 0:
                result.append(default_value)
            else:
                result.append(func(df_subset))

            start, end = start + 1, end + 1

        # If we haven't filled up the results fully, we append default value. If the offset is negative, we
        # put them at the start - otherwise we put them at the end
        if len(result) < len(self.obj):
            default_values = [default_value for _ in range(len(self.obj) - len(result))]
            # If the offset is negative
            if self.offset < 0:
                result = default_values + result
            else:
                result = result + default_values

        return pd.Series(result, index=self.obj.index)


import ast
from collections import Counter
from copy import copy
from typing import Any, Dict, Iterable, List, Optional, Tuple
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
from mitosheet.ai.ai_utils import fix_up_missing_imports, get_code_string_from_last_expression, replace_last_instance_in_string

from io import StringIO
from contextlib import redirect_stdout

from mitosheet.errors import MitoError, make_column_exists_error, make_exec_error
from mitosheet.state import DATAFRAME_SOURCE_AI, State
from mitosheet.step_performers.dataframe_steps.dataframe_delete import \
    delete_dataframe_from_state
from mitosheet.types import (AITransformFrontendResult, ColumnHeader, ColumnID,
                             DataframeReconData, ModifiedDataframeReconData)

def is_df_changed(old: pd.DataFrame, new: pd.DataFrame) -> bool:
    try:
        assert_frame_equal(old, new, check_names=False)
        return False
    except AssertionError:
        return True

def exec_for_recon(code: str, original_df_map: Dict[str, pd.DataFrame]) -> DataframeReconData:
    """
    Given some Python code, and a list of previously defined dataframes, this function:
    1. Gets all the newly defined dataframes
    2. Gets all the dataframes that have been modified
    3. Gets the value of the last line of code, if that is a Python expression (e.g. not defining a variable)

    It does all of this while only execing the code a single time, which helps performance.

    To find newly defined dataframes, we just save the local variables before the exec runs, 
    and then find any new dataframes in locals after the exec runs.

    To find the modified dataframes, we capture any dataframe names that are referenced in the code
    as potentially modified and add them as a local. This then allows us to get out the new values 
    of that dataframe, and compare them to the old one to see if they really changed.

    Finially, getting the value of the last line requires parsing the Python AST, checking the
    last expressio, then rebuilding it into a string (aka so we can handle multiple lines) - 
    and then saving it in fake variable that we can then access again through the locals.
    """
    if len(code) == 0:
        return {
            'created_dataframes': {},
            'deleted_dataframes': [],
            'modified_dataframes': {},
            'last_line_expression_value': None,
            'prints': ''
        }

    df_map = {df_name: df.copy(deep=True) for df_name, df in original_df_map.items()}
    locals_before = copy(locals())
    try:
        ast_before = ast.parse(code)
    except SyntaxError as e:
        raise make_exec_error(e)

    try:
        last_expression = ast_before.body[-1]
    except:
        raise make_exec_error(Exception('No code was generated'))
    
    if not isinstance(last_expression, ast.Expr):
        has_last_line_expression_value = False
    else:
        has_last_line_expression_value = True
        # (The type ignore is b/c we know this has a value b/c of above check, mypy is not smart enough)
        last_expression_string: str = get_code_string_from_last_expression(code) #type: ignore
        code = replace_last_instance_in_string(code, last_expression_string, f'FAKE_VAR_NAME = {last_expression_string}')
    
    potentially_modified_df_names = [
        df_name for df_name in original_df_map if 
        df_name in code
    ]

    for df_name in potentially_modified_df_names:
        locals()[df_name] = df_map[df_name]

    # If there are lambda expressions in the code, then we need to also 
    # inject the pandas and numpy modules in as globals -- as for some
    # reason the code inside of a lambda doesn't have access to them
    globals = {}
    if 'lambda' in code:
        globals['pd'] = pd
        globals['np'] = np

    # Capture the output as well
    output_string_io = StringIO()

    try:
        with redirect_stdout(output_string_io):
            exec(code, globals, locals())
    except Exception as e:
        raise make_exec_error(e)
        
    # We make a copy of locals, as the local variables redefine themselves in 
    # list comprehensions... sometimes. I don't get what is happening here, but it works
    new_locals = locals()

    created_dataframes: Dict[str, pd.DataFrame] = {
        name: value for name, value in new_locals.items()
        if name not in locals_before and name != 'locals_before' and name != 'FAKE_VAR_NAME'
        and isinstance(value, pd.DataFrame) and name not in original_df_map
    }

    deleted_dataframes = [
        name for name in df_map if name not in new_locals and name in potentially_modified_df_names
    ]

    modified_dataframes = {
        name: value for name, value in new_locals.items()
        if isinstance(value, pd.DataFrame) and name in potentially_modified_df_names
        and is_df_changed(original_df_map[name], value)
    }

    if has_last_line_expression_value:
        last_line_expression_value = locals()['FAKE_VAR_NAME']
    else:
        last_line_expression_value = None

    return {
        'created_dataframes': created_dataframes,
        'deleted_dataframes': deleted_dataframes,
        'modified_dataframes': modified_dataframes,
        'last_line_expression_value': last_line_expression_value,
        'prints': output_string_io.getvalue()
    }

PANDAS_HAS_NA = hasattr(pd, 'NA')

def is_null_column_header_in_column_headers(column_header: ColumnHeader, column_headers: Iterable[ColumnHeader]) -> bool:

    # First, check for NA. Notably, we need to check that NA is
    # is a attribute of pandas, as it's only available in pandas 1.0+
    if PANDAS_HAS_NA and column_header is pd.NA:
        return any(c is pd.NA for c in column_headers)
        
    # Then, check for NaT
    if column_header is pd.NaT:
        return any(c is pd.NaT for c in column_headers)
    
    # Then, check for NaN
    if column_header is np.nan:
        return any(c is np.nan for c in column_headers)
    
    # Then, check None
    if column_header is None and None in column_headers:
        return True
    
    return False

def is_possibly_null_column_header_in_column_headers_with_no_nans(column_header: ColumnHeader, column_headers_with_no_nans: Iterable[ColumnHeader]) -> bool:
    """
    Checks if the column header is in the list of column headers, taking special care to handle if the column
    header is null. Notably, assumes that the column_headers have no null values.
    """
    if not pd.isna(column_header):
        return column_header in column_headers_with_no_nans
    
    return is_null_column_header_in_column_headers(column_header, column_headers_with_no_nans)


def get_added_column_headers(old_column_headers: List[ColumnHeader], new_column_headers: Iterable[ColumnHeader]) -> List[ColumnHeader]:

    old_non_null = list(filter(lambda ch: not pd.isna(ch), old_column_headers))
    new_non_null = list(filter(lambda ch: not pd.isna(ch), new_column_headers))
    added_non_null = list(filter(lambda ch: ch not in old_non_null, new_non_null))

    new_null = list(filter(lambda ch: pd.isna(ch), new_column_headers))
    old_null = list(filter(lambda ch: pd.isna(ch), old_column_headers))
    added_null = list(filter(lambda ch: not is_null_column_header_in_column_headers(ch, old_null), new_null))

    return added_non_null + added_null

def get_shared_column_headers(old_column_headers: List[ColumnHeader], new_column_headers: Iterable[ColumnHeader]) -> List[ColumnHeader]:

    old_non_null = list(filter(lambda ch: not pd.isna(ch), old_column_headers))
    new_non_null = list(filter(lambda ch: not pd.isna(ch), new_column_headers))
    shared_non_null = list(filter(lambda ch: ch in old_non_null, new_non_null))

    new_null = list(filter(lambda ch: pd.isna(ch), new_column_headers))
    old_null = list(filter(lambda ch: pd.isna(ch), old_column_headers))
    shared_null = list(filter(lambda ch: is_null_column_header_in_column_headers(ch, old_null), new_null))

    return shared_non_null + shared_null


def get_modified_dataframe_recon_data(old_df: pd.DataFrame, new_df: pd.DataFrame) -> ModifiedDataframeReconData:
    """
    Given a dataframe and a modified dataframe, this function tries to figure out what has happened
    to column headers dataframe. Specifically, because our state maps column headers to do others based on column
    id, we need to track which columns are added, which are removed, and which are renamed.
    """

    old_columns = old_df.columns.to_list()
    new_columns = new_df.columns.to_list()

    # If there are duplicated columns in the new dataframe, we throw an error
    # as we don't support this elsewhere in the codebase
    if len(new_columns) != len(set(new_columns)):
        raise MitoError(
            'duplicated_column_headers_error', 
            'Duplicated Column Headers', 
            'There are duplicated columns in the new dataframe',
            error_modal=False
        )

    old_df_head = old_df.head(5)
    new_df_head = new_df.head(5)

    rows_added_or_removed = len(old_df) != len(new_df)

    # First, preserving the order, we remove any columns that are in both the old
    # and the new dataframe
    old_columns_without_shared = get_added_column_headers(new_columns, old_columns)
    new_columns_without_shared = get_added_column_headers(old_columns, new_columns)
    
    # Then, we look through to find any columns that have been simply renamed - simply
    # by comparing to see of column are identical between the two values. We do this 
    # just by checking the first 5 values of the dataframe, before doing a direct comparison
    renamed_columns: Dict[ColumnHeader, ColumnHeader] = {}
    for old_ch in old_columns_without_shared:
        old_column = old_df_head[old_ch]
        for new_ch in new_columns_without_shared:
            new_column = new_df_head[new_ch]
            if old_column.equals(new_column) and new_ch not in renamed_columns.values():
                renamed_columns[old_ch] = new_ch

    added_columns = [ch for ch in new_columns_without_shared if not is_possibly_null_column_header_in_column_headers_with_no_nans(ch, renamed_columns.values())]
    removed_columns = [ch for ch in old_columns_without_shared if not is_possibly_null_column_header_in_column_headers_with_no_nans(ch, renamed_columns)]

    shared_columns = get_shared_column_headers(old_columns, new_columns)

    if not rows_added_or_removed:
        modified_columns = [ch for ch in shared_columns if not old_df[ch].equals(new_df[ch])]
    else:
        # If rows were added or removed, then we don't want to detect every column as having changed
        # and instead we'd just like to report the row changes. As such, we only compare the rows not added or removed
        # NOTE: this is not perfect, as you may have modified columns and removed rows in one go -- but this 
        # is ok for most of what we see
        try:
            if len(old_df) < len(new_df):
                df1 = old_df
                df2 = new_df.loc[old_df.index]
            else:
                df1 = old_df.loc[new_df.index]
                df2 = new_df

            modified_columns = [ch for ch in shared_columns if not df1[ch].equals(df2[ch])]
        except IndexError:
            modified_columns = [ch for ch in shared_columns if not old_df[ch].equals(new_df[ch])]

    return {
        'column_recon': {
            'created_columns': added_columns,
            'deleted_columns': removed_columns,
            'modified_columns': modified_columns,
            'renamed_columns': renamed_columns,
        },
        'num_added_or_removed_rows': len(new_df) - len(old_df)
    }

def delete_column_id_from_state_metadata(
    state: State,
    sheet_index: int,
    column_id: ColumnID,
) -> State:
    
    # Update all the state variables removing this column from the state
    del state.column_formulas[sheet_index][column_id]
    # TODO: do we want to remove the formulas
    if column_id in state.df_formats[sheet_index]['columns']:
        del state.df_formats[sheet_index]['columns'][column_id]

    # Clean up the IDs
    state.column_ids.delete_column_id(sheet_index, column_id)
        
    return state

def update_state_by_reconing_dataframes(
        state: State, 
        sheet_index: int, 
        old_df: pd.DataFrame,
        new_df: pd.DataFrame,
        column_headers_to_column_ids: Optional[Dict[ColumnHeader, ColumnID]]=None
    ) -> Tuple[State, ModifiedDataframeReconData]:
    """
    This function is the work-horse for modified dataframes. It compares the old dataframe at the index 
    to the new dataframe, and then updates the state accordingly -- making sure all the metadata is correct.

    This includes: handling deleted columns, added columns, renamed columns, and modified columns.
    """
    # Check there aren't any duplicated columns in the new dataframe
    c = Counter(new_df.columns)
    most_common = c.most_common(1)
    for ch, count in most_common:
        if count > 1:
            raise make_column_exists_error(ch)

    modified_dataframe_recon = get_modified_dataframe_recon_data(old_df, new_df)

    # Add new columns to the state
    if len(modified_dataframe_recon['column_recon']['created_columns']) > 0:
        state.add_columns_to_state(
            sheet_index, 
            modified_dataframe_recon['column_recon']['created_columns'],
            column_headers_to_column_ids=column_headers_to_column_ids
        )

    # Delete removed columns from the state
    deleted_column_ids = state.column_ids.get_column_ids_by_headers(sheet_index, modified_dataframe_recon['column_recon']['deleted_columns'])
    for column_id in deleted_column_ids:
        delete_column_id_from_state_metadata(state, sheet_index, column_id)

    # Rename renamed columns in the state
    for old_ch, new_ch in modified_dataframe_recon['column_recon']['renamed_columns'].items():
        column_id = state.column_ids.get_column_id_by_header(sheet_index, old_ch)
        state.column_ids.set_column_header(sheet_index, column_id, new_ch)

    # Then, actually set the dataframe
    state.dfs[sheet_index] = new_df

    return state, modified_dataframe_recon


def exec_and_get_new_state_and_result(state: State, code: str) -> Tuple[State, Optional[Any], AITransformFrontendResult]:

    # Fix up the code, so we can ensure that we execute it properly
    code = fix_up_missing_imports(code)

    # Make deep copies of all dataframes here, so we can manipulate them without fear
    # TODO: in the future, we could just do the modified ones
    new_state = state.copy(deep_sheet_indexes=list(range(len(state.dfs))))

    df_map = {df_name: df for df_name, df in zip(new_state.df_names, new_state.dfs)}
    recon_data = exec_for_recon(code, df_map)

    # For all of the added dataframes, we just add them to the state
    for df_name, df in recon_data['created_dataframes'].items():
        new_state.add_df_to_state(df, DATAFRAME_SOURCE_AI, df_name=df_name)

    # For deleted dataframes, we remove them from the state
    for df_name in recon_data['deleted_dataframes']:
        delete_dataframe_from_state(new_state, new_state.df_names.index(df_name))
    
    # For modified dataframes, we update all the column variables
    modified_dataframes_recons: Dict[str, ModifiedDataframeReconData] = {}
    for df_name, new_df in recon_data['modified_dataframes'].items():
        sheet_index = new_state.df_names.index(df_name)
        new_state, modified_dataframes_recon = update_state_by_reconing_dataframes(new_state, sheet_index, new_state.dfs[sheet_index], new_df)
        modified_dataframes_recons[df_name] = modified_dataframes_recon

    # For the last value, if is a dataframe, then add it to the state as well -- unless this dataframe
    # is a _newly_ created dataframe that is already given a name
    last_line_expression_value = recon_data['last_line_expression_value']
    last_line_expression_code = get_code_string_from_last_expression(code)
    last_line_expression_is_previously_created_dataframe = last_line_expression_code in recon_data['created_dataframes']
    if (isinstance(last_line_expression_value, pd.DataFrame) or isinstance(last_line_expression_value, pd.Series)) and not last_line_expression_is_previously_created_dataframe:

        # If we get a series, we turn it into a dataframe for the user
        if isinstance(last_line_expression_value, pd.Series):
            last_line_expression_value = pd.DataFrame(last_line_expression_value, index=last_line_expression_value.index)

        new_state.add_df_to_state(last_line_expression_value, DATAFRAME_SOURCE_AI)
        # We also need to add this to the list of created dataframes, as we didn't know it's name till now
        recon_data['created_dataframes'][new_state.df_names[-1]] = last_line_expression_value

    # If the last line value is a primitive, we return it as a result for the frontend
    result_last_line_value = None
    if isinstance(last_line_expression_value, str) or isinstance(last_line_expression_value, bool) \
        or isinstance(last_line_expression_value, int) or isinstance(last_line_expression_value, float) \
            or isinstance(last_line_expression_value, np.number):
        result_last_line_value = last_line_expression_value

    frontend_result: AITransformFrontendResult = {
        'last_line_value': result_last_line_value,
        'created_dataframe_names': list(recon_data['created_dataframes'].keys()),
        'deleted_dataframe_names': recon_data['deleted_dataframes'],
        'modified_dataframes_recons': modified_dataframes_recons,
        'prints': recon_data['prints']
    }

    return (new_state, recon_data['last_line_expression_value'], frontend_result)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List
from mitosheet.saved_analyses.step_upgraders.utils_column_header_to_column_id import \
    replace_headers_with_id


def upgrade_reorder_column_1_to_2(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Moves to using column id instead of column header.

    OLD: {
        'step_version': 1, 
        'step_type': "reorder_column",
        'params': {
            sheet_index: 0,
            column_header: _header_,
            new_column_index: 2,
        }
    }

    NEW: {
        'step_version': 2, 
        'step_type': "reorder_column",
        'params': {
            sheet_index: 0,
            column_id: _id_,
            new_column_index: 1
        }
    }
    """
    params = step['params']
    params = replace_headers_with_id(params, 'column_header', 'column_id')

    return [{
        "step_version": 2, 
        "step_type": "reorder_column", 
        "params": params
    }] + later_steps



from typing import List, TypeVar, Dict

T = TypeVar('T')

def deduplicate_array(my_list: List[T]) -> List:
    seen: Dict[T, T] = {}
    new_list = [seen.setdefault(x, x) for x in my_list if x not in seen]
    return new_list



from datetime import datetime, timedelta
from typing import Optional, Union


def get_million_identifier_in_string(string: str) -> Union[str, None]:
    """
    Given a string, returns the million identifier in it. 
    Returns '' if none exist. 
    """
    million_identifiers = ["Million", 'Mil', 'M', 'million', 'mil', 'm']
    # So that we return the biggest matching element
    million_identifiers = list(sorted(million_identifiers, key=len, reverse=True))

    for identifier in million_identifiers:
        if identifier in string:
            return identifier

    return None


def get_billion_identifier_in_string(string: str) -> Union[str, None]:
    """
    Given a string, returns the billion identifier in it. 
    Returns '' if none exist. 
    """

    billion_identifiers = ["Billion", 'Bil', 'B', 'billion', 'bil', 'b']
    # So that we return the biggest matching element
    billion_identifiers = list(sorted(billion_identifiers, key=len, reverse=True))

    for identifier in billion_identifiers:
        if identifier in string:
            return identifier

    return None


def cast_string_to_float(
        s: str, 
    ) -> Optional[float]:
    """
    NOTE: The approach for this function is to start as a string, and then we try
    and turn each element into a number 1-by-1. We attempt to handle:
    1. All basic integers / floats.
    2. Dollar signs. 
    3. Parenthases to denote negative numbers, per accounting conventions. 
    4. Commas in the string (e.g. 123,456 => 123456). NOTE: We handle the European conventions here too, 
        where commas may represent a decimal. However, this is impossible to make perfect, as 123,123 and
        123,123 are 123123 and 123.123 in America and Europe respectively. We treat this as American for 
        now. 
    5. Million or Billion identifier in the string
    6. Percentage sign in the string
    As these are all heuristics, we do our best. We also try to perform this conversion
    optimistically, as to run as quickly as possible.
    """
    try:
        # Try to handle case 1, optimistically
        return float(s)
    except:
        # Get rid of whitespace at the ends
        s = s.strip()

        is_negative = False
        if s.startswith('-'):
            s = s[1:]
            is_negative = True

        # Handle 2, if it exist
        if s.startswith('$'):
            s = s[1:]

        # Handle 3, if it exists
        if s.startswith('(') and s.endswith(')'):
            s = s[1:-1]
            is_negative = True

        # Handle case 4, if it's an issue
        if ',' in s:
            # We try and figure out if it's a european or american comma usage, by seeing
            # what happens at the end of the string
            last_comma_index = s.rfind(',')

            # If there is no period, and the the string after the last comma is anything other
            # than 3 characters long, than we take this as a European, and turn the comma into a .
            if '.' not in s and last_comma_index != len(s) - 4:
                s = s.replace(',', '.')
            else:
                # Otherwise, we treat this as American
                s = s.replace(',', '')

        # Handle case 5, if there is a million or billion identifier in the number
        million_identifer = get_million_identifier_in_string(s)
        billion_identifier = get_billion_identifier_in_string(s)
        multiplier = 1.0

        if million_identifer != None:
            multiplier = 1000000
            s = s.replace(million_identifer, '') # type: ignore

        if billion_identifier != None:
            multiplier = 1000000000
            s = s.replace(billion_identifier, '') # type: ignore

        # Handle case 6, if there is a percentage sign in the number
        if s.endswith('%'):
            s = s[:-1]
            multiplier = multiplier / 100

        try:
            return float(s) * (-1 if is_negative else 1) * multiplier
        except:
            return None
        
def cast_to_float(unknown: Union[str, int, float, bool, datetime, timedelta]) -> Optional[float]:
    if isinstance(unknown, str):
        return cast_string_to_float(unknown)
    elif isinstance(unknown, int):
        return float(unknown)
    elif isinstance(unknown, float):
        return unknown
    elif isinstance(unknown, bool):
        return float(unknown)

    return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import pandas as pd
import pytest
from mitosheet.enterprise.mito_config import MitoConfig
from mitosheet.types import FORMULA_ENTIRE_COLUMN_TYPE

from mitosheet.utils import get_new_id
from mitosheet.errors import MitoError
from mitosheet.steps_manager import StepsManager
from mitosheet.tests.test_utils import create_mito_wrapper_with_data
from mitosheet.column_headers import get_column_header_id


def test_create_steps_manager():
    df1 = pd.DataFrame(data={'A': [123]})
    df2 = pd.DataFrame(data={'A': [123]})

    mito_config = MitoConfig()

    steps_manager = StepsManager([df1, df2], mito_config)
    assert steps_manager.curr_step_idx == 0
    assert steps_manager.curr_step.step_type == 'initialize'
    assert steps_manager.curr_step.column_formulas == [{'A': []}, {'A': []}]
    assert steps_manager.curr_step.dfs[0].equals(df1)
    assert steps_manager.curr_step.dfs[1].equals(df2)
    assert steps_manager.mito_config.mito_config_dict == mito_config.mito_config_dict

# We assume only column A exists
CELL_EDIT_ERRORS = [
    ('=HI()', 'unsupported_function_error'),
    ('=UPPER(HI())', 'unsupported_function_error'),
    ('=UPPER(HI())', 'unsupported_function_error'),
    ('=C', 'no_column_error'),
    ('=C + D', 'no_column_error'),
    ('=ABCDEFG', 'no_column_error'),
    ('=UPPER(C)', 'no_column_error'),
    ('=UPPER(A, 100)', 'function_error'),
    ('=UPPER(LOWER(A, 100))', 'function_error')
]

@pytest.mark.parametrize("formula,error_type", CELL_EDIT_ERRORS)
def test_steps_manager_cell_edit_errors(formula,error_type):
    mito = create_mito_wrapper_with_data([123])
    mito.add_column(0, 'B')
    with pytest.raises(MitoError) as e_info:
        mito.mito_backend.steps_manager.handle_edit_event({
            'event': 'edit_event',
            'id': get_new_id(),
            'type': 'set_column_formula_edit',
            'step_id': get_new_id(),
            'params': {
                'sheet_index': 0,
                'column_id': get_column_header_id('B'),
                'formula_label': 0,
                'index_labels_formula_is_applied_to': {'type': FORMULA_ENTIRE_COLUMN_TYPE},
                'old_formula': '=0',
                'new_formula': formula
            }
        })
    assert e_info.value.type_ == error_type


def test_overwrites_step_valid():
    mito = create_mito_wrapper_with_data([1, 2, 3])
    mito.add_column(0, 'B')

    mito.mito_backend.receive_message({
        'event': 'edit_event',
        'id': get_new_id(),
        'type': 'add_column_edit',
        'step_id': mito.mito_backend.steps_manager.curr_step.step_id,
        'params': {
            'sheet_index': 0,
            'column_header': 'C',
            'column_header_index': 1
        }
    })

    assert len(mito.steps_including_skipped) == 3
    assert mito.dfs[0].equals(pd.DataFrame(data={'A': [1, 2, 3], 'C': [0, 0, 0]}))


def test_failed_overwrite_rolls_back_to_previous_state():
    mito = create_mito_wrapper_with_data([1, 2, 3])
    mito.add_column(0, 'B')

    mito.mito_backend.receive_message({
        'event': 'edit_event',
        'id': get_new_id(),
        'type': 'add_column_edit',
        'step_id': mito.mito_backend.steps_manager.curr_step.step_id,
        'params': {
            'sheet_index': 1,
            'column_header': 'C',
            'column_header_index': 2
        }
    })

    assert len(mito.steps_including_skipped) == 2
    assert mito.dfs[0].equals(pd.DataFrame(data={'A': [1, 2, 3], 'B': [0, 0, 0]}))





#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

import os
from mitosheet.enterprise.license_key import encode_date_to_license
from mitosheet.enterprise.mito_config import (
    DEFAULT_MITO_CONFIG_SUPPORT_EMAIL, 
    MEC_VERSION_KEYS,
    MITO_CONFIG_ANALYTICS_URL,
    MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL, 
    MITO_CONFIG_LOG_SERVER_URL,
    MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL, 
    MITO_CONFIG_CODE_SNIPPETS, 
    MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL, 
    MITO_CONFIG_CODE_SNIPPETS_URL, 
    MITO_CONFIG_CODE_SNIPPETS_VERSION,
    MITO_CONFIG_CUSTOM_IMPORTERS_PATH,
    MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH,
    MITO_CONFIG_DISABLE_TOURS,
    MITO_CONFIG_ENTERPRISE,
    MITO_CONFIG_ENTERPRISE_TEMP_LICENSE,
    MITO_CONFIG_LLM_URL, 
    MITO_CONFIG_SUPPORT_EMAIL, 
    MITO_CONFIG_VERSION, 
    MITO_CONFIG_SUPPORT_EMAIL, 
    MITO_CONFIG_VERSION, 
    MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT,
    MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT,
    MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION,
    MITO_CONFIG_FEATURE_DISPLAY_SCHEDULING,
    MITO_CONFIG_FEATURE_DISPLAY_CODE_OPTIONS,
    MITO_CONFIG_FEATURE_TELEMETRY,
    MITO_CONFIG_PRO,
    MitoConfig
)
from mitosheet.tests.test_utils import create_mito_wrapper


def delete_env_var_if_exists(env_var: str) -> None: 
    """
    Deletes the environment variable only if it exists to avoid errors. Helpful for testing.
    """
    if os.environ.get(env_var) is not None:
        del os.environ[env_var]

def delete_all_mito_config_environment_variables() -> None:
    """
    Deletes all of the MITO CONFIG environment variables that are used in the MEC
    """
    lists_of_vars = MEC_VERSION_KEYS.values()
    for list_of_vars in lists_of_vars:
        for var in list_of_vars:
            delete_env_var_if_exists(var)

def test_keys_did_not_change():
    # We must not change these keys so we can still read old 
    # mito configs that users have not upgraded to the most 
    # recent config version. If we don't preserve these keys, 
    # we won't know which environment variables to read.
    assert MITO_CONFIG_VERSION == 'MITO_CONFIG_VERSION'
    assert MITO_CONFIG_SUPPORT_EMAIL == 'MITO_CONFIG_SUPPORT_EMAIL'
    assert MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL == 'MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL'

def test_none_works():
    # Delete the environmnet variables so we can test the none condition
    delete_all_mito_config_environment_variables()

    mito_config = MitoConfig()
    mito_config_dict = mito_config.mito_config_dict
    assert mito_config_dict == {
        MITO_CONFIG_VERSION: '2',
        MITO_CONFIG_SUPPORT_EMAIL: DEFAULT_MITO_CONFIG_SUPPORT_EMAIL,
        MITO_CONFIG_DISABLE_TOURS: False,
        MITO_CONFIG_CODE_SNIPPETS: None,
        MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT: False,
        MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT: True,
        MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION: True,
        MITO_CONFIG_FEATURE_DISPLAY_SCHEDULING: True,
        MITO_CONFIG_FEATURE_DISPLAY_CODE_OPTIONS: True,
        MITO_CONFIG_LLM_URL: None,
        MITO_CONFIG_ANALYTICS_URL: None,
        MITO_CONFIG_LOG_SERVER_URL: None,
        MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL: None, 
        MITO_CONFIG_FEATURE_TELEMETRY: True,
        MITO_CONFIG_PRO: False,
        MITO_CONFIG_ENTERPRISE: False,
        MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH: None,
        MITO_CONFIG_CUSTOM_IMPORTERS_PATH: None
    }

def test_none_config_version_is_string():
    # Delete the environmnet variables so we can test the none condition
    delete_all_mito_config_environment_variables()

    mito_config = MitoConfig()
    mito_config_dict = mito_config.mito_config_dict
    assert isinstance(mito_config_dict[MITO_CONFIG_VERSION], str)

def test_version_2_works():
    # Set environment variables
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_SUPPORT_EMAIL] = "aaron@sagacollab.com"
    os.environ[MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL] = "jake@sagacollab.com"
    os.environ[MITO_CONFIG_CODE_SNIPPETS_URL] = "url"
    os.environ[MITO_CONFIG_DISABLE_TOURS] = "True"
    os.environ[MITO_CONFIG_CODE_SNIPPETS_VERSION] = "1"

    # Test reading environment variables works properly
    mito_config = MitoConfig()
    assert mito_config.mito_config_dict == {
        MITO_CONFIG_VERSION: '2',
        MITO_CONFIG_SUPPORT_EMAIL: 'aaron@sagacollab.com',
        MITO_CONFIG_DISABLE_TOURS: True,
        MITO_CONFIG_CODE_SNIPPETS: {
            MITO_CONFIG_CODE_SNIPPETS_VERSION : '1',
            MITO_CONFIG_CODE_SNIPPETS_URL: 'url',
            MITO_CONFIG_CODE_SNIPPETS_SUPPORT_EMAIL: "jake@sagacollab.com"
        },
        MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT: False,
        MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT: True,
        MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION: True,
        MITO_CONFIG_FEATURE_DISPLAY_SCHEDULING: True, 
        MITO_CONFIG_FEATURE_DISPLAY_CODE_OPTIONS: True,
        MITO_CONFIG_LLM_URL: None,
        MITO_CONFIG_ANALYTICS_URL: None,
        MITO_CONFIG_LOG_SERVER_URL: None,
        MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL: None, 
        MITO_CONFIG_FEATURE_TELEMETRY: True,
        MITO_CONFIG_PRO: False,
        MITO_CONFIG_ENTERPRISE: False,
        MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH: None,
        MITO_CONFIG_CUSTOM_IMPORTERS_PATH: None
    }

    # Delete the environmnet variables for the next test
    delete_all_mito_config_environment_variables()

def test_mito_config_update_version_1_to_2():
    # Set environment variables
    os.environ[MITO_CONFIG_VERSION] = "1"
    os.environ[MITO_CONFIG_SUPPORT_EMAIL] = "aaron@sagacollab.com"
    
    # Test reading environment variables works properly
    mito_config = MitoConfig()
    assert mito_config.mito_config_dict == {
        MITO_CONFIG_VERSION: '2',
        MITO_CONFIG_SUPPORT_EMAIL: 'aaron@sagacollab.com',
        MITO_CONFIG_DISABLE_TOURS: False,
        MITO_CONFIG_CODE_SNIPPETS: None,
        MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT: False,
        MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT: True,
        MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION: True,
        MITO_CONFIG_FEATURE_DISPLAY_SCHEDULING: True, 
        MITO_CONFIG_FEATURE_DISPLAY_CODE_OPTIONS: True,
        MITO_CONFIG_LLM_URL: None,
        MITO_CONFIG_ANALYTICS_URL: None,
        MITO_CONFIG_LOG_SERVER_URL: None,
        MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL: None, 
        MITO_CONFIG_FEATURE_TELEMETRY: True,
        MITO_CONFIG_PRO: False,
        MITO_CONFIG_ENTERPRISE: False,
        MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH: None,
        MITO_CONFIG_CUSTOM_IMPORTERS_PATH: None
    }    

    # Delete the environmnet variables for the next test
    delete_all_mito_config_environment_variables()

def test_mito_config_enable_snowflake_import():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_SUPPORT_EMAIL] = "aaron@sagacollab.com"
    os.environ[MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT] = "True"
    
    mito_config = MitoConfig()
    assert mito_config.mito_config_dict == {
        MITO_CONFIG_VERSION: '2',
        MITO_CONFIG_SUPPORT_EMAIL: 'aaron@sagacollab.com',
        MITO_CONFIG_DISABLE_TOURS: False,
        MITO_CONFIG_CODE_SNIPPETS: None,
        MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT: True,
        MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT: True,
        MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION: True,
        MITO_CONFIG_FEATURE_DISPLAY_SCHEDULING: True, 
        MITO_CONFIG_FEATURE_DISPLAY_CODE_OPTIONS: True,
        MITO_CONFIG_LLM_URL: None,
        MITO_CONFIG_ANALYTICS_URL: None,
        MITO_CONFIG_LOG_SERVER_URL: None,
        MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL: None, 
        MITO_CONFIG_FEATURE_TELEMETRY: True,
        MITO_CONFIG_PRO: False,
        MITO_CONFIG_ENTERPRISE: False,
        MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH: None,
        MITO_CONFIG_CUSTOM_IMPORTERS_PATH: None
    }    

    delete_all_mito_config_environment_variables()

def test_mito_config_dont_display_features():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_SUPPORT_EMAIL] = "aaron@sagacollab.com"
    os.environ[MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT] = "False"
    os.environ[MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION] = "F"
    os.environ[MITO_CONFIG_FEATURE_DISPLAY_SCHEDULING] = "0"
    os.environ[MITO_CONFIG_FEATURE_DISPLAY_CODE_OPTIONS] = "false"
    
    mito_config = MitoConfig()
    assert mito_config.mito_config_dict == {
        MITO_CONFIG_VERSION: '2',
        MITO_CONFIG_SUPPORT_EMAIL: 'aaron@sagacollab.com',
        MITO_CONFIG_DISABLE_TOURS: False,
        MITO_CONFIG_CODE_SNIPPETS: None,
        MITO_CONFIG_FEATURE_ENABLE_SNOWFLAKE_IMPORT: False,
        MITO_CONFIG_FEATURE_DISPLAY_SNOWFLAKE_IMPORT: False,
        MITO_CONFIG_FEATURE_DISPLAY_AI_TRANSFORMATION: False,
        MITO_CONFIG_FEATURE_DISPLAY_SCHEDULING: False, 
        MITO_CONFIG_FEATURE_DISPLAY_CODE_OPTIONS: False, 
        MITO_CONFIG_LLM_URL: None,
        MITO_CONFIG_ANALYTICS_URL: None,
        MITO_CONFIG_LOG_SERVER_URL: None,
        MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL: None, 
        MITO_CONFIG_FEATURE_TELEMETRY: True,
        MITO_CONFIG_PRO: False,
        MITO_CONFIG_ENTERPRISE: False,
        MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH: None,
        MITO_CONFIG_CUSTOM_IMPORTERS_PATH: None
    }    

    delete_all_mito_config_environment_variables()

def test_mito_config_disable_telemetry():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_FEATURE_TELEMETRY] = "False"

    from mitosheet.telemetry.telemetry_utils import telemetry_turned_on
    assert not telemetry_turned_on()

    delete_all_mito_config_environment_variables()


def test_mito_config_enable_pro():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_PRO] = "True"

    from mitosheet.user import is_pro
    assert is_pro()

    delete_all_mito_config_environment_variables()


def test_mito_config_enable_enterprise():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_ENTERPRISE] = "True"

    from mitosheet.user import is_enterprise
    assert is_enterprise()

    delete_all_mito_config_environment_variables()

def test_mito_config_enable_enterprise_date():
    
    # Get tomorrow as a date
    from datetime import date, timedelta
    tomorrow = date.today() + timedelta(days=1)

    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_ENTERPRISE_TEMP_LICENSE] = encode_date_to_license(tomorrow)

    from mitosheet.user import is_enterprise
    assert is_enterprise()

    delete_all_mito_config_environment_variables()

def test_mito_config_disable_enterprise_date():
    
    # Get tomorrow as a date
    from datetime import date, timedelta
    yesterday = date.today() + timedelta(days=-1)
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_ENTERPRISE_TEMP_LICENSE] = encode_date_to_license(yesterday)

    from mitosheet.user import is_enterprise
    assert not is_enterprise()

    delete_all_mito_config_environment_variables()

def test_mito_config_custom_sheet_functions_path():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH] = "./folder/file.py"

    mito_config = MitoConfig()
    mito_config = mito_config.mito_config_dict
    mito_config[MITO_CONFIG_VERSION] == '2'
    mito_config[MITO_CONFIG_CUSTOM_SHEET_FUNCTIONS_PATH] == "./folder/file.py"

    delete_all_mito_config_environment_variables()

def test_mito_config_custom_importers_path():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_CUSTOM_IMPORTERS_PATH] = "./folder/file.py"

    mito_config = MitoConfig()
    mito_config = mito_config.mito_config_dict
    mito_config[MITO_CONFIG_VERSION] == '2'
    mito_config[MITO_CONFIG_CUSTOM_IMPORTERS_PATH] == "./folder/file.py"

    delete_all_mito_config_environment_variables()

def test_mito_config_log_server_url():
    
    os.environ[MITO_CONFIG_VERSION] = "2"
    os.environ[MITO_CONFIG_LOG_SERVER_URL] =  "url"
    os.environ[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] = "15"

    mito_config = MitoConfig()
    mito_config = mito_config.mito_config_dict
    mito_config[MITO_CONFIG_VERSION] == '2'
    mito_config[MITO_CONFIG_CUSTOM_IMPORTERS_PATH] == "url"
    mito_config[MITO_CONFIG_LOG_SERVER_BATCH_INTERVAL] == 15

    # Test that the env var are used to create the mito_log_uploader
    mito = create_mito_wrapper()
    steps_manager = mito.mito_backend.steps_manager
    steps_manager.mito_log_uploader.log_url == "url"
    steps_manager.mito_log_uploader.base_log_interval == 15

    delete_all_mito_config_environment_variables()





#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import os

MITO_CONFIG_KEY_HOME_FOLDER = 'MITO_CONFIG_HOME_FOLDER'
if MITO_CONFIG_KEY_HOME_FOLDER in os.environ:
    HOME_FOLDER = os.path.expanduser(os.environ[MITO_CONFIG_KEY_HOME_FOLDER])
else:
    HOME_FOLDER = os.path.expanduser('~')

# Where all global .mito files are stored
MITO_FOLDER = os.path.join(HOME_FOLDER, ".mito")



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import List, Tuple, Any, Dict
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.types import UserDefinedFunctionParamType
from mitosheet.state import State
from mitosheet.step_performers.utils.user_defined_function_utils import get_transpiled_user_defined_function_params

class UserDefinedEditCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, edit_name: str, sheet_index: int, user_defined_function_params: Dict[str, Tuple[UserDefinedFunctionParamType, Any, Any]]):
        super().__init__(prev_state)
        self.edit_name = edit_name
        self.sheet_index = sheet_index
        self.user_defined_function_params = user_defined_function_params

        self.df_name = prev_state.df_names[self.sheet_index]

    def get_display_name(self) -> str:
        return 'User Defined Edit'
    
    def get_description_comment(self) -> str:
        return f"Edited sheet {self.df_name} with {self.edit_name}"

    def get_code(self) -> Tuple[List[str], List[str]]:
        code = f"{self.df_name} = {self.edit_name}({get_transpiled_user_defined_function_params(self.user_defined_function_params)})"
        return [code], []

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index] 
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk


class NoOpCodeChunk(CodeChunk):
    """A code chunk that generates no code and should always be optimized out"""

    def get_display_name(self) -> str:
        return "NoOp"
    
    def get_description_comment(self) -> str:
        return "No comment"

    def get_code(self) -> Tuple[List[str], List[str]]:
        return [], []

    def combine_right(self, other_code_chunk: CodeChunk) -> Optional[CodeChunk]:
        # The empty code chunk always overwrites itself with the right code chunk
        return other_code_chunk


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import json
import os
from typing import Any, Dict

import pandas as pd
from mitosheet.code_chunks.step_performers.import_steps.simple_import_code_chunk import DEFAULT_DECIMAL, DEFAULT_DELIMITER, DEFAULT_ENCODING, DEFAULT_SKIPROWS
from mitosheet.step_performers.import_steps.simple_import import read_csv_get_delimiter_and_encoding
from mitosheet.types import StepsManagerType


def get_csv_files_metadata(params: Dict[str, Any], steps_manager: StepsManagerType) -> Dict[str, Any]:
    """
    Given a list of 'file_names' that should be CSV files,
    this returns our guesses for delimeters and encodings
    for these files, as well as other default parameters that we 
    don't try to guess
    """
    file_names = params['file_names']

    delimeters = []
    encodings = []
    decimals = []
    skiprows = []
    for file_name in file_names:
        try:
            _, delimiter, encoding = read_csv_get_delimiter_and_encoding(file_name)
            delimeters.append(delimiter)
            encodings.append(encoding)
        except:
            # The default values displayed in the UI
            delimeters.append(DEFAULT_DELIMITER)
            encodings.append(DEFAULT_ENCODING)

        # We don't have a good way to guess these params, so we always use the defaults
        decimals.append(DEFAULT_DECIMAL)
        skiprows.append(DEFAULT_SKIPROWS)

    return {
        'delimeters': delimeters,
        'encodings': encodings,
        'decimals': decimals,
        'skiprows': skiprows 
    }


